#! /usr/bin/python

''' helper functions '''
import numpy as np
import pandas as pd
import copy, re, os, geojson
from scipy import sparse
from scipy.special import gammaln, psi, logsumexp, expit, logit
from sklearn.preprocessing import normalize
import scipy.stats
import scipy.optimize
import sklearn.cluster

from matplotlib.patches import Rectangle
import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
from matplotlib.path import Path

from sklearn.utils import check_random_state
from sklearn.decomposition._online_lda_fast import (
    _dirichlet_expectation_1d, _dirichlet_expectation_2d,
)

def dirichlet_expectation(alpha):
    """
    For a vector theta ~ Dir(alpha), computes E[log(theta)] given alpha.
    """
    assert alpha.min() > 0, "Expecting positive Dirichlet parameters"
    if (len(alpha.shape) == 1):
        return( _dirichlet_expectation_2d(alpha.reshape((-1, 1))) )
    return _dirichlet_expectation_2d(alpha)

def init_latent_vars(model, n_features, dtype=np.float64, gamma = None, ):
    """Initialize latent variables."""
    model.random_state_ = check_random_state(model.random_state)
    model.n_batch_iter_ = 1
    model.n_iter_ = 0
    if model.doc_topic_prior is None:
        model.doc_topic_prior_ = 1.0 / model.n_components
    else:
        model.doc_topic_prior_ = model.doc_topic_prior
    if model.topic_word_prior is None:
        model.topic_word_prior_ = 1.0 / model.n_components
    else:
        model.topic_word_prior_ = model.topic_word_prior

    init_gamma = 100.0
    init_var = 1.0 / init_gamma
    # In the literature, this is called `lambda`
    if gamma is None:
        model.components_ = model.random_state_.gamma(
            init_gamma, init_var, (model.n_components, n_features)
        ).astype(dtype, copy=False)
    else:
        assert gamma.shape == (model.n_components, n_features)
        model.components_ = gamma.astype(dtype, copy=False)

    # In the literature, this is `exp(E[log(beta)])`
    model.exp_dirichlet_component_ = np.exp(
        _dirichlet_expectation_2d(model.components_)
    )

def pg_mean(b,c=0):
    if np.isscalar(c) and c == 0:
        return b/4
    if not np.isscalar(c):
        if np.isscalar(b):
            b = np.ones(c.shape[0]) * b
        v = b/4
        indx = (c != 0)
        v[indx] = b[indx]/2/c[indx] * np.tanh(c[indx]/2)
        return v
    else:
        return (b/2/c * np.tanh(c/2))

def real_to_sb(mtx):
    assert len(mtx.shape) == 2, "Invalid matrix"
    n, K = mtx.shape
    phi = expit(mtx)
    s = phi[:, 0]
    for k in range(1, K-1):
        phi[:, k] = phi[:, k] * (1. - s)
        s +=  phi[:, k]
    phi[:, K - 1] = 1. - s
    phi = np.clip(phi, 1e-8, 1.-1e-8)
    phi = normalize(phi, norm='l1', axis=1)
    return phi

def sb_to_real(phi):
    assert len(phi.shape) == 2, "Invalid matrix"
    phi = np.clip(phi, 1e-8, 1.-1e-8)
    phi = normalize(phi, norm='l1', axis=1) * (1.-1e-6)
    n, K = phi.shape
    mtx = np.zeros((n, K))
    mtx[:, 0] = logit(phi[:, 0])
    s = phi[:, 0]
    for k in range(1, K):
        mtx[:, k] = logit(phi[:, k] / (1. - s) )
        s += phi[:, k]
    return mtx

def logsumexp_csr(X):
    assert sparse.issparse(X), "logsumexp_csr: invalid input"
    X = X.tocsr()
    result = np.zeros(len(X.data))
    for i in range(X.shape[0]):
        result[X.indptr[i]:X.indptr[i+1]] =\
            logsumexp(X.data[X.indptr[i]:X.indptr[i+1]])
    return result

def gen_even_slices(n, n_packs):
    start = 0
    if n_packs < 1:
        raise ValueError("gen_even_slices got n_packs=%s, must be >=1" % n_packs)
    for pack_num in range(n_packs):
        this_n = n // n_packs
        if pack_num < n % n_packs:
            this_n += 1
        if this_n > 0:
            end = start + this_n
            yield np.arange(start, end)
            start = end

def gen_even_slices_from_list(vec, n_packs):
    start = 0
    vec = np.array(vec)
    n = len(vec)
    if n_packs < 1:
        raise ValueError("gen_even_slices got n_packs=%s, must be >=1" % n_packs)
    for pack_num in range(n_packs):
        this_n = n // n_packs
        if pack_num < n % n_packs:
            this_n += 1
        if this_n > 0:
            end = start + this_n
            yield vec[start:end]
            start = end

def gen_slices_from_list(vec, bsize):
    start = 0
    vec = np.array(vec)
    n = len(vec)
    while start < n:
        end = start + bsize
        if end > n - bsize//2:
            end = n
        yield vec[start:end]
        start = end

def get_string_with_integer_suff(in_array):
    out = []
    for u in in_array:
        v = re.match('^[A-Za-z]*_*(\d+)$', u.strip())
        if v:
            out.append(v.group(0))
    return out

def get_integer_suff_from_string(in_array):
    out = []
    for u in in_array:
        v = re.match('^[A-Za-z]*_*(\d+)$', u.strip())
        if v:
            out.append(v.group(1))
    return out

def extract_polygons_from_json(gj):
    if isinstance(gj, str):
        assert os.path.exists(gj), f"Invalid file\n{gj}"
        gj = geojson.load(open(gj,'rb'))
    assert gj['type'] == 'FeatureCollection', "Invalid geojson file"
    vertices = []
    for geom in gj['features']:
        if geom['geometry']['type'] == 'Polygon':
            poly = geom['geometry']['coordinates']
            vertices.append(np.array(poly).squeeze())
        elif geom['geometry']['type'] == 'MultiPolygon':
            for poly in geom['geometry']['coordinates']:
                vertices.append(np.array(poly).squeeze())
    return vertices

def svg_parse_list(path):
    commands = {'M': (Path.MOVETO,),'L': (Path.LINETO,),
                'Q': (Path.CURVE3,)*2,'C': (Path.CURVE4,)*3,
                'Z': (Path.CLOSEPOLY,) }
    vertices = []
    codes = []
    for v in path:
        cmd = v[0].upper()
        if len(v) > 1:
            points = [float(x) for x in v[1:]]
            points = np.array(points).reshape((len(points)//2,2))
            vertices.extend(points.tolist())
            codes.extend(commands[cmd])
    return codes, vertices

def match_factors(mtx1, mtx2, c1, n, cmap, mode='beta'):
    """
    Harmonize a pair of LDA results
    """
    k1 = mtx1.shape[0]
    k2 = mtx2.shape[0]
    assert mtx1.shape[1] == mtx2.shape[1], "Invalid input matrices"

    # Try to match two sets of factors
    Q = np.zeros((k1, k2))
    if mode == 'beta':
        beta1 = mtx1 / mtx1.sum(axis = 1).reshape((-1, 1))
        beta2 = mtx2 / mtx2.sum(axis = 1).reshape((-1, 1))
        for k in range(k1):
            q,r = scipy.optimize.nnls(beta2.transpose(), beta1[k, :].transpose())
            Q[k, ] = q
            indx2 = [(-Q[:,k]).argsort() for k in range(k2)]
            indx1 = [(-Q[k,:]).argsort() for k in range(k1)]
    else:
        for k in range(k1):
            q,r = scipy.optimize.nnls(mtx2.transpose(), mtx1[k, :].transpose())
            Q[k, ] = q
            indx2 = [(-Q[:,k]).argsort() for k in range(k2)]
            indx1 = [(-Q[k,:]).argsort() for k in range(k1)]
            Q = Q * mtx2.sum(axis = 1).reshape((1, -1))
            Q = Q / (Q.sum(axis = 1).reshape((-1, 1)))

    model = sklearn.cluster.SpectralBiclustering(n_clusters=(k1,k1), method="bistochastic").fit(Q+1e-5)
    prio = (-Q.sum(axis = 0)).argsort() # sorted from most consistent to less consistent
    step = n//(k1*2)

    dup = [1] * len(c1)
    c1_ref = copy.copy(c1)
    c2 = [-1]*k2
    for i,k in enumerate(prio):
        c2[k] = c1_ref[indx2[k][0]]
        even = dup[indx2[k][0]] % 2
        odd  = dup[indx2[k][0]] + even
        c1_ref[indx2[k][0]] += int(step * (-1)**(even) * (1-1/odd))
        dup[indx2[k][0]] += 1

    indx1 = np.array(c1).argsort()
    v1 = mtx1.sum(axis = 1)
    v1 = v1/v1.sum()
    v1 = v1[indx1]

    indx2 = np.array(c2).argsort()
    v2 = mtx2.sum(axis = 1)
    v2 = v2/v2.sum()
    v2 = v2[indx2]

    st = []
    ed = []
    st_y = []
    ed_y = []
    cl = []

    st_y += [0] + list(np.cumsum(v1)[:-1])
    ed_y += list(np.cumsum(v1))
    cl += [c1[x] for x in indx1]

    offset = 0
    st += [k1+offset]*k1
    ed += [k1+offset]*k1

    st_y += [0] + list(np.cumsum(v2)[:-1])
    ed_y += list(np.cumsum(v2))
    cl += [c2[x] for x in indx2]

    st += [k2-offset]*k2
    ed += [k2-offset]*k2

    df = pd.DataFrame({'xst':st,'yst':st_y,'xed':ed,'yed':ed_y,'Color':cl})

    mid1 = []
    mid2 = []
    v = np.cumsum(v1)
    mid1+=list(v - v1*0.5)
    v = np.cumsum(v2)
    mid2+=list(v - v2*0.5)

    xst = []
    xed = []
    yst = []
    yed = []
    sl = []
    cl = []
    offset = 0.3
    # for ii,i in enumerate(indx1):
    for ii,i in enumerate(indx1):
        for jj,j in enumerate(indx2):
            if Q[i,j] > 1e-3:
                xst.append(k2-offset)
                xed.append(k1+offset)
                yst.append(mid2[jj])
                yed.append(mid1[ii])
                sl.append(Q[i, j])
                cl.append(c1[i])
    df_arrow = pd.DataFrame({'xst':xst,'yst':yst,'xed':xed,'yed':yed,'Weight':sl,'Color':cl})
    return df, df_arrow, c2

def plot_colortable(colors, title, sort_colors=True, ncols=4, dpi = 80,\
                    cell_width = 212, cell_height = 22,\
                    title_fontsize=24, text_fontsize=24,\
                    swatch_width = 48, margin = 12, topmargin = 40):
    # Sort colors by hue, saturation, value and name.
    if sort_colors is True:
        by_hsv = sorted((tuple(mcolors.rgb_to_hsv(mcolors.to_rgb(color))),
                         name)
                        for name, color in colors.items())
        names = [name for hsv, name in by_hsv]
    else:
        names = list(colors)

    n = len(names)
    nrows = n // ncols + int(n % ncols > 0)

    width = cell_width * ncols + 2 * margin
    height = cell_height * nrows + margin + topmargin


    fig, ax = plt.subplots(figsize=(width / dpi, height / dpi), dpi=dpi)
    fig.subplots_adjust(margin/width, margin/height,
                        (width-margin)/width, (height-topmargin)/height)
    ax.set_xlim(0, cell_width * ncols)
    ax.set_ylim(cell_height * (nrows-0.5), -cell_height/2.)
    ax.yaxis.set_visible(False)
    ax.xaxis.set_visible(False)
    ax.set_axis_off()
    ax.set_title(title, fontsize=title_fontsize, loc="left", pad=10)

    for i, name in enumerate(names):
        row = i % nrows
        col = i // nrows
        y = row * cell_height

        swatch_start_x = cell_width * col
        text_pos_x = cell_width * col + swatch_width + 7

        ax.text(text_pos_x, y, name, fontsize=text_fontsize,
                horizontalalignment='left',
                verticalalignment='center')

        ax.add_patch(
            Rectangle(xy=(swatch_start_x, y-9), width=swatch_width,
                      height=18, facecolor=colors[name], edgecolor='0.7')
        )

    return fig

def scale_to_prob(W, H):
    uk = H.sum(axis = 1).reshape((-1, 1))
    beta = H / uk
    sn = W @ uk
    theta = (W * uk.T)/sn
    return theta, beta

def read_ct_from_solo_barcodes_tsv(file, key, chunksize=500000, mu_scale=-1):
    ct_idx = "gn,gt,spl,unspl,ambig".split(",").index(key)
    reader = pd.read_csv(gzip.open(f,'rb'), sep='\t', usecols=[1,5,6,7], index_col=0, names=["unit","X","Y",key], dtype={'X':int, 'Y':int}, chunksize=chunksize)
    for chunk in reader:
        if mu_scale > 0:
            chunk.X /= mu_scale
            chunk.Y /= mu_scale
        chunk[key]=chunk[key].map(lambda x : x.split(',')[ct_idx]).astype(int)
        yield chunk

def make_mtx_from_dge(file, min_ct_per_feature = 50, min_ct_per_unit = 100, feature_white_list = None, feature_list = None, unit = "random_index", key = "gn", epoch=1, epoch_id_length=2, return_df = False, feature_list_force = False):
    df = pd.DataFrame()
    epoch_id_list = set()
    for chunk in pd.read_csv(file, sep='\t', usecols = [unit,'X','Y','gene',key], dtype={unit:str}, chunksize=500000):
        unit_list = chunk[unit].str[:epoch_id_length].unique()
        i = 0
        while len(epoch_id_list) < epoch and i < len(unit_list):
            epoch_id_list.add(unit_list[i])
            i += 1
        if len(epoch_id_list) >= epoch and unit_list[-1] not in epoch_id_list:
            pat = '^' + '|'.join(list(epoch_id_list))
            df = pd.concat([df, chunk[chunk[unit].str.contains(pat) & chunk[key].ge(1)] ])
            break
        df = pd.concat([df, chunk[chunk[key].ge(1)] ])
    epoch0 = df[unit].iloc[0][:epoch_id_length]
    one_pass = df[unit].str[:epoch_id_length].eq(epoch0)
    if feature_list is not None:
        feature = pd.DataFrame({"gene": feature_list})
        ct = df[one_pass].groupby(by=['gene']).agg({key:"sum"}).reset_index()
        if feature_list_force is False:
            ct.drop(index = ct[ct[key].lt(min_ct_per_feature)].index, inplace=True)
            feature = feature.merge(right = ct, on = 'gene', how = 'inner')
        else:
            feature = feature.merge(right = ct, on = 'gene', how = 'left')
            feature.fillna(0, inplace=True)
    else:
        feature = df[one_pass].groupby(by=['gene']).agg({key:"sum"}).reset_index()
        feature_white_list = set() if feature_white_list is None else set(feature_white_list)
        feature = feature.loc[feature[key].ge(min_ct_per_feature ) | feature.gene.isin(feature_white_list), :]
    M = len(feature)
    feature.index = np.arange(M)
    ft_dict = {x:i for i,x in enumerate(feature.gene)}
    df.drop(index = df.index[~df.gene.isin(ft_dict)], inplace=True)

    brc = df.groupby(by = unit).agg({key:"sum"}).reset_index()
    brc.drop(index = brc.index[brc[key].lt(min_ct_per_unit)], inplace=True)
    N = brc.shape[0]
    brc.index = np.arange(N)
    bc_dict = {x:i for i,x in enumerate(brc[unit])}
    brc['j'] = brc[unit].map(bc_dict)
    df.drop(index = df.index[~df[unit].isin(bc_dict)], inplace=True)
    df['j'] = df[unit].map(bc_dict)
    brc["epoch"] = brc[unit].str[:epoch_id_length]
    df.drop(columns = unit, inplace=True)
    brc = brc.merge(right = df[['j','X','Y' ]].drop_duplicates(subset='j'), on = 'j', how = 'left')

    mtx = sparse.coo_array((df[key].values, (df.j.values, df.gene.map(ft_dict))), shape=(N, M)).tocsr()

    feature["Weight"] = mtx.sum(axis = 0)
    feature.Weight = feature.Weight * 1. / feature.Weight.sum()

    if return_df:
        return df, feature, brc, mtx, ft_dict, bc_dict
    else:
        return feature, brc, mtx, ft_dict, bc_dict


def chisq(k,info,total_k,total_umi):
    res = []
    if total_k <= 0:
        return res
    for name, v in info.iterrows():
        if v[k] <= 0:
            continue
        tab=np.zeros((2,2))
        tab[0,0]=v[str(k)]
        tab[0,1]=v["gene_tot"]-tab[0,0]
        tab[1,0]=total_k-tab[0,0]
        tab[1,1]=total_umi-total_k-v["gene_tot"]+tab[0,0]
        fd=tab[0,0]/total_k/tab[0,1]*(total_umi-total_k)
        if fd < 1:
            continue
        tab = np.around(tab, 0).astype(int) + 1
        chi2, p, dof, ex = scipy.stats.chi2_contingency(tab, correction=False)
        res.append([name,k,chi2,p,fd,v["gene_tot"]])
    return res



import sys, io, os, gzip, glob, copy, re, time, warnings, pickle, argparse, logging
from collections import defaultdict,Counter
import numpy as np
import pandas as pd
from datetime import datetime

from scipy.sparse import *
from sklearn.utils import check_random_state
from sklearn.preprocessing import normalize
from joblib import Parallel, delayed
from sklearn.decomposition import LatentDirichletAllocation as LDA

sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from ficture.utils.utilt import make_mtx_from_dge, init_latent_vars
from ficture.loaders.unit_loader import UnitLoader

def init_model_from_pseudobulk(_args):

    parser = argparse.ArgumentParser()
    parser.add_argument('--input', type=str, help='')
    parser.add_argument('--output', type=str, help='')
    parser.add_argument('--model', type=str, help='')
    parser.add_argument('--feature', type=str, default = '', help='') # if provided the output will only contains these features
    parser.add_argument('--key', type=str, default='gn', help='')
    parser.add_argument('--thread', type=int, default=1, help='')
    parser.add_argument('--seed', type=int, default=-1, help='')

    parser.add_argument('--scale_model_rel', type=float, default=0, help='Scale the total magnitude of the input pseudobulk model relative to the total count of one epoch of the input. Set to <=0 to use the input model as is.')
    parser.add_argument('--epoch', type=int, default=0, help='Epoch for model training, 0 for using the input model as is')
    parser.add_argument('--alpha', type=float, default=1, help='')
    parser.add_argument('--tau', type=int, default=9, help='')
    parser.add_argument('--kappa', type=float, default=0.7, help='')

    parser.add_argument('--unit_label', default = 'random_index', type=str, help='Which column to use as unit identifier')
    parser.add_argument('--feature_label', default = "gene", type=str, help='Which column to use as feature identifier')
    parser.add_argument('--epoch_id_length', type=int, default=2, help='')
    parser.add_argument('--min_ct_per_unit', type=int, default=50, help='')
    parser.add_argument('--min_ct_per_feature', type=int, default=50, help='')
    parser.add_argument('--reorder_factors', action='store_true', help='')
    parser.add_argument('--debug', action='store_true', help='')

    args = parser.parse_args(_args)
    if len(_args) == 0:
        parser.print_help()
        return

    logging.basicConfig(level= getattr(logging, "INFO", None))

    seed = int(args.seed)
    if seed <= 0:
        seed = int(datetime.now().timestamp()) % 2147483648
    rng = check_random_state(seed)

    key = args.key
    thread = args.thread
    unit_key = args.unit_label.lower()
    gene_key = args.feature_label.lower()
    b_size = 512
    min_ct_model_feature = 50

    # Read model
    model_mtx = pd.read_csv(args.model, sep='\t', index_col = 0)
    model_gene_sum = model_mtx.sum(axis = 1)
    model_mtx.drop(index = model_gene_sum[model_gene_sum<min_ct_model_feature].index, inplace=True )
    factor_header = list(model_mtx.columns)
    feature_model =list(model_mtx.index)
    logging.info(f"Read model matrix {model_mtx.shape}")

    # Read input
    feature_list = None
    if os.path.isfile(args.feature):
        feature_list = pd.read_csv(args.feature, sep='\t', dtype={gene_key:str})[gene_key].tolist()
        feature_list = list(set(feature_list))
    feature, brc, mtx_org, ft_dict, bc_dict = make_mtx_from_dge(args.input,\
        min_ct_per_unit = args.min_ct_per_unit, \
        min_ct_per_feature = args.min_ct_per_feature, \
        feature_list = feature_list, \
        unit = args.unit_label, key = key)
    N, M0 = mtx_org.shape
    logging.info(f"Read data with {N} units, {M0} features")

    # Intersection of features
    feature_model_idx = [ft_dict[x] for x in set(feature_model) & ft_dict.keys()]
    feature_unused_idx = [ft_dict[x] for x in ft_dict.keys() - set(feature_model)]
    if len(feature_model_idx) <= 2:
        sys.exit("ERROR: too few features in common between the model and the input data")
    feature_model = feature.loc[feature_model_idx, "gene"].values
    model_mtx = model_mtx.loc[feature_model, :].values + 0.1
    M, K = model_mtx.shape
    N = mtx_org.shape[0]
    logging.info(f"Use {M} overlapping features for model fitting")

    mtx_org = mtx_org.tocsc()
    M1 = len(feature_unused_idx)
    mtx_unused = mtx_org[:, feature_unused_idx].tocsr()
    mtx_org  = mtx_org[:, feature_model_idx].tocsr()

    if args.scale_model_rel > 0:
        raw_sum = mtx_org.sum()
        model_mtx = model_mtx * (args.scale_model_rel * raw_sum / model_mtx.sum())
    model = LDA(n_components=K, learning_method='online', batch_size=b_size, total_samples = N, learning_offset = args.tau, learning_decay = args.kappa, doc_topic_prior = args.alpha, n_jobs = args.thread, verbose = 0, random_state=seed )
    init_latent_vars(model, n_features = M, gamma = model_mtx.T)

    # Update model
    e = 0
    while e < args.epoch:
        idx = rng.permutation(N)
        _ = model.partial_fit(mtx_org[idx,:])
        e += 1

    # Fit all units
    oheader = ["unit",key,"x","y","topK","topP"]+factor_header
    dtp = {'topK':int,key:int,"unit":str}
    dtp.update({x:float for x in ['topP']+factor_header})
    theta = model.transform(mtx_org)
    post_count = np.array(theta.T @ mtx_org)
    feature_output = feature_model
    if M1 > 0:
        post_count_unused = np.array(theta.T @ mtx_unused)
        post_count = np.concatenate((post_count, post_count_unused), axis = 1)
        feature_output = feature.loc[feature_model_idx + feature_unused_idx, "gene"].values

    # Relabel factors
    if args.reorder_factors:
        weight = post_count.sum(axis=1)
        ordered_k = np.argsort(weight)[::-1]
        model.components_ = model.components_[ordered_k,:]
        model.exp_dirichlet_component_ = model.exp_dirichlet_component_[ordered_k,:]
        theta = theta[:, ordered_k]
        post_count = post_count[ordered_k,:]
        factor_header = [factor_header[x] for x in ordered_k]

    if args.epoch > 0:
        out_f = args.output + ".model_matrix.tsv.gz"
        pd.concat([pd.DataFrame({gene_key: feature_model}),\
            pd.DataFrame(model.components_.T, columns = factor_header, dtype='float64')], axis = 1\
        ).to_csv(out_f, sep='\t', index=False, float_format='%.4e', compression={"method":"gzip"})
        model.feature_names_in_ = feature_model
        pickle.dump(model, open(args.output + ".model.p", 'wb'))

    # Unit assignment
    brc.rename(columns = {'j':'unit', 'X':'x', 'Y':'y'}, inplace = True)
    brc['topK'] = np.argmax(theta, axis = 1).astype(int)
    brc['topP'] = np.max(theta, axis = 1)
    brc = pd.concat((brc, pd.DataFrame(theta, columns = factor_header )), axis = 1)
    brc = brc.astype(dtp)
    res_f = args.output+".fit_result.tsv.gz"
    brc[oheader].to_csv(res_f, sep='\t', float_format="%.4e", index=False, header=True, compression={"method":"gzip"})
    logging.info(f"Result file {res_f}")

    # Posterior count
    out_f = args.output+".posterior.count.tsv.gz"
    pd.concat([pd.DataFrame({gene_key: feature_output}),\
            pd.DataFrame(post_count.T, columns = factor_header, dtype='float64')],\
        axis = 1).to_csv(out_f, sep='\t', index=False, float_format='%.2f', compression={"method":"gzip"})

if __name__ == "__main__":
    init_model_from_pseudobulk(sys.argv[1:])



### Stream in pixel level data, assign minibatch label
### global-randomize-local-contiguous
### Output has the same columns as input with an extra column (1st) being the minibatch index

import sys, os, argparse, gzip, logging, copy, time
import subprocess as sp
import numpy as np
import pandas as pd
import random as rng

def make_spatial_minibatch(_args):

    parser = argparse.ArgumentParser(prog = "make_spatial_minibatch")

    # Innput and output info
    parser.add_argument('--input', type=str, help='')
    parser.add_argument('--output', type=str, help='')

    # Basic parameter
    parser.add_argument('--major_axis', type=str, default="Y", help='X or Y')
    parser.add_argument('--major_axis_max', type=float, default=-1, help='')
    parser.add_argument('--batch_size', type=float, default=500, help='Length of the side (um) of square minibatches')
    parser.add_argument('--batch_buff', type=float, default=50, help='Overlap between minibatches')
    parser.add_argument('--mu_scale', type=float, default=26.67, help='Coordinate to um translate')
    parser.add_argument('--min_pixel', type=int, default=100, help='Just to avoid non-tissue regions with isolated pixels')
    parser.add_argument('--seed', type=float, default=-1, help='')

    # Specify a region to work on
    parser.add_argument('--regions', type=str, default="", help='A file containing region intervals, same as that used by tabix -R (e.g. tab delimited, lane start end)')
    parser.add_argument('--region', nargs='*', type=str, default=[], help='lane:Ystart-Yend (Y axis in barcode coordinate unit), separate by space if multiple regions')
    parser.add_argument('--region_um', nargs='*', type=str, default=[], help='lane:Ystart-Yend (Y axis in um), separate by space if multiple regions')

    args = parser.parse_args(_args)
    if len(_args) == 0:
        parser.print_help()
        return

    logging.basicConfig(level= getattr(logging, "INFO", None))

    if not os.path.exists(args.input):
        sys.exit("ERROR: cannot find input file")

    r_seed = args.seed
    if r_seed <= 0:
        r_seed = time.time()
    rng.seed(r_seed)
    logging.info(f"Random seed {r_seed}")

    ### Basic parameterse
    batch_size = int(args.batch_size * args.mu_scale)
    batch_buff = int(args.batch_buff * args.mu_scale)
    batch_step = batch_size - batch_buff

    ### Input pixel info (input has to have correct header)
    with gzip.open(args.input, 'rt') as rf:
        input_header=rf.readline().strip().split('\t')
    for i,x in enumerate(input_header):
        if x in ["x", "y"]:
            input_header[i] = input_header[i].upper()
    dty = {x:float for x in ['X', 'Y']}

    ### Streaming input (Is this safe?)
    cmd = []
    if os.path.exists(args.regions):
        cmd = ["tabix", args.input, "-R", args.regions]
    elif len(args.region) > 0:
        cmd = ["tabix", args.input] + args.region
    elif len(args.region_um) > 0:
        reg_list = []
        for v in args.region_um:
            if ":" not in v or "-" not in v:
                continue
            l = v.split(':')[0]
            st, ed = v.split(':')[1].split('-')
            st = str(int(float(st) * args.mu_scale) )
            ed = str(int(float(ed) * args.mu_scale) )
            reg_list.append(l+':'+'-'.join([st,ed]) )
        cmd = ["tabix", args.input] + reg_list
    if len(cmd) == 0:
        p0 = sp.Popen(["gzip", "-cd", args.input], stdout=sp.PIPE)
        process = sp.Popen(["tail", "-n", "+2"], stdin=p0.stdout, stdout=sp.PIPE)
    else:
        process = sp.Popen(cmd, stdout=sp.PIPE, stderr=sp.STDOUT)

    ### Group pixels into minibatches
    output_header = copy.copy(input_header)
    output_header.insert(1, "random_index")
    print(output_header)
    if args.output.endswith(".gz"):
        with gzip.open(args.output, 'wt') as wf:
            _=wf.write('\t'.join(output_header) + '\n')
    else:
        with open(args.output, 'w') as wf:
            _=wf.write('\t'.join(output_header) + '\n')

    df = pd.DataFrame()
    nbatch=0
    reader = pd.read_csv(process.stdout,sep='\t',chunksize=1000000,\
        names=input_header, dtype=dty)
    end_of_file = False

    while not end_of_file:
        chunk = next(reader, None)
        if chunk is None:
            end_of_file = True
            if len(df) == 0:
                break
        else:
            chunk["random_index"] = -1
            chunk.random_index = chunk.random_index.astype(int)
            df = pd.concat([df, chunk])
        x_min, y_min = df[["X","Y"]].values.min(axis = 0)
        x_max, y_max = df[["X","Y"]].values.max(axis = 0)
        x_range = x_max - x_min
        y_range = y_max - y_min
        logging.info(f"Read blocks of pixels: {x_range/args.mu_scale:.2f} x {y_range/args.mu_scale:.2f}")
        if not end_of_file:
            if x_range < batch_size or y_range < batch_size:
                continue
            if args.major_axis_max > 0:
                if args.major_axis == "X" and x_max > args.major_axis_max - batch_size / 2:
                    continue
                if args.major_axis == "Y" and y_max > args.major_axis_max - batch_size / 2:
                    continue
        else: # end of file, and the last batch is too small potentially caused by the batch size almost divides the axis lenght
            if (args.major_axis == "X" and x_range <= batch_buff) or (args.major_axis == "Y" and y_range <= batch_buff):
                break
        x_grd_st = np.arange(x_min, x_max-batch_size/2+1, batch_step)
        if len(x_grd_st) == 0:
            x_grd_st = np.array([x_min-1])
        x_grd_ed = [x + batch_size for x in x_grd_st]
        y_grd_st = np.arange(y_min, y_max-batch_size/2+1, batch_step)
        if len(y_grd_st) == 0:
            y_grd_st = np.array([y_min-1])
        y_grd_ed = [x + batch_size for x in y_grd_st]
        x_grd_st[0]  = x_min - 1
        y_grd_st[0]  = y_min - 1
        x_grd_ed[-1] = x_max
        y_grd_ed[-1] = y_max

        for it_x in range(len(x_grd_st)):
            for it_y in range(len(y_grd_st)):
                xst = x_grd_st[it_x]
                xed = x_grd_ed[it_x]
                yst = y_grd_st[it_y]
                yed = y_grd_ed[it_y]
                indx = (df.X > xst) & (df.X <= xed) & (df.Y > yst) & (df.Y <= yed)
                if sum(indx) < args.min_pixel:
                    continue
                df.loc[indx, "random_index"] = rng.randint(1, sys.maxsize//100)
                ### Output
                xst /= args.mu_scale
                xed /= args.mu_scale
                yst /= args.mu_scale
                yed /= args.mu_scale
                logging.info(f"Output region ({xed-xst:.2f}, {yed-yst:.2f}) ({xst:.1f}, {xed:.1f}) x ({yst:.1f}, {yed:.1f})")
                df.loc[indx, output_header].to_csv(args.output, mode='a', sep='\t', index=False, header=False, float_format='%.2f')
                nbatch += 1

        ### Leftover
        if end_of_file:
            break
        mj_max = df.loc[:, args.major_axis].max()
        left_indx = (df.random_index < 0) |\
                    (df[args.major_axis] > mj_max-batch_buff)
        df = df.loc[left_indx, :]
        df["random_index"] = -1
        w  = df[args.major_axis].max() - df[args.major_axis].min()
        w0 = (df.random_index < 0).sum()
        logging.info(f"Left over size {df.shape[0]} ({w0}, {w:.2f})")


if __name__ == "__main__":
    make_spatial_minibatch(sys.argv[1:])



import re, os, sys
#sys.path.append(os.path.dirname(os.path.abspath(__file__)))

class minimake:
    def __init__(self):
        # Dictionary to store the target as keys and a tuple of source files and commands as values
        self.targets = {}

    def add_target(self, target_file, source_files, commands):
        """
        Add a target with its dependencies and the commands to build it.

        :param target_file: String, the name of the target file
        :param source_files: List of strings, the source files the target depends on
        :param commands: List of strings, the commands needed to build the target
        """
        if target_file in self.targets:
            raise ValueError(f"Target {target_file} already exists.")
        self.targets[target_file] = (source_files, commands)

    def write_makefile(self, output_filename):
        """
        Writes a Makefile with all added targets.

        :param output_filename: String, the path to the output Makefile
        """
        with open(output_filename, 'w') as file:
            file.write("# Makefile generated by minimake\n\n")
            file.write("all: " + " ".join(self.targets.keys()) + "\n\n")
            for target, (sources, commands) in self.targets.items():
                # Write the target and its dependencies
                file.write(f"{target}: {' '.join(sources)}\n")
                # Write each command prefixed by a tab character
                for command in commands:
                    file.write(f"\t{command}\n")
                file.write("\n")  # add a newline for readability between targets

def create_source_files():
    with open("main.c", "w") as f:
        f.write(
            """#include "utils.h"
int main() {
    return print_hello();
}
"""
        )

    with open("utils.c", "w") as f:
        f.write(
            """#include <stdio.h>
int print_hello() {
    printf("Hello, world!\\n");
    return 0;
}
"""
        )

    with open("utils.h", "w") as f:
        f.write(
            """#ifndef UTILS_H
#define UTILS_H
int print_hello();
#endif
"""
        )

# Example usage
if __name__ == "__main__":
    create_source_files()
    mm = minimake()
    mm.add_target("program.exe", ["main.o", "utils.o"], ["gcc -o program.exe main.o utils.o"])
    mm.add_target("main.o", ["main.c", "utils.h"], ["gcc -c main.c"])
    mm.add_target("utils.o", ["utils.c", "utils.h"], ["gcc -c utils.c"])
    mm.write_makefile("Makefile")


import sys, os, copy, gzip, logging
import pickle, argparse
import numpy as np
import pandas as pd

from scipy.sparse import *
import sklearn.neighbors
import sklearn.preprocessing
from sklearn.decomposition import LatentDirichletAllocation as LDA

from ficture.loaders.unit_loader import UnitLoader
from ficture.utils.utilt import init_latent_vars

def lda(_args):

    parser = argparse.ArgumentParser(prog = "lda")
    parser.add_argument('--input', type=str, help='')
    parser.add_argument('--output', '--output_pref', type=str, help='')
    parser.add_argument('--feature', type=str, help='')
    parser.add_argument('--unit_label', default = 'random_index', type=str, help='Which column to use as unit identifier')
    parser.add_argument('--unit_attr', type=str, nargs='+', default=[], help='')
    parser.add_argument('--feature_label', default = "gene", type=str, help='Which column to use as feature identifier')
    parser.add_argument('--key', default = 'count', type=str, help='')
    parser.add_argument('--train_on', default = '', type=str, help='')
    parser.add_argument('--log', default = '', type=str, help='files to write log to')
    parser.add_argument('--shift_log_transform', action='store_true')
    parser.add_argument('--fix_scaling', type = float, default = -1,)

    parser.add_argument('--nFactor', type=int, default=10, help='')
    parser.add_argument('--minibatch_size', type=int, default=512, help='')
    parser.add_argument('--min_ct_per_feature', type=int, default=1, help='')
    parser.add_argument('--min_ct_per_unit', type=int, default=20, help='')
    parser.add_argument('--thread', type=int, default=1, help='')
    parser.add_argument('--epoch', type=int, default=1, help='How many times to loop through the full data')
    parser.add_argument('--epoch_id_length', type=int, default=-1, help='')
    parser.add_argument('--prior', type=str, default='', help="Dirichlet parameters for the global parameter beta (factor x gene)")
    parser.add_argument('--rescale_prior', type = float, default = -1,)
    parser.add_argument('--alpha', type=float, default=1, help='')
    parser.add_argument('--tau', type=int, default=9, help='')
    parser.add_argument('--kappa', type=float, default=0.7, help='')
    parser.add_argument('--N', type=float, default=1e4, help='')
    parser.add_argument('--seed', type=int, default=-1, help='')
    parser.add_argument('--debug', action='store_true')
    parser.add_argument('--verbose', action='store_true')
    parser.add_argument('--overwrite', action='store_true')

    args = parser.parse_args(_args)
    if len(_args) == 0:
        parser.print_help()
        return

    if args.log != '':
        try:
            logging.basicConfig(filename=args.log, filemode='a', encoding='utf-8', level=logging.INFO)
        except:
            logging.basicConfig(level= getattr(logging, "INFO", None))
    else:
        logging.basicConfig(level= getattr(logging, "INFO", None))

    seed = args.seed
    if seed <= 0:
        seed = int(datetime.now().timestamp() )

    unit_attr = [x.lower() for x in args.unit_attr]
    key = args.key.lower()
    train_on = args.train_on.lower()
    unit_key = args.unit_label.lower()
    gene_key = args.feature_label.lower()
    if train_on == '':
        train_on = key
    adt = {unit_key:str, gene_key:str, key:int, train_on:int}
    adt.update({x:str for x in unit_attr})
    print(unit_attr)

    ### Basic parameterse
    b_size = args.minibatch_size
    K = args.nFactor

    ### Input
    # Required columns: unit ID, gene, key
    required_header = [unit_key,gene_key,train_on]
    if not os.path.exists(args.input):
        sys.exit("ERROR: cannot find input file.")
    with gzip.open(args.input, 'rt') as rf:
        header = rf.readline().strip().split('\t')
    header = [x.lower() for x in header]
    for x in required_header:
        if x not in header:
            sys.exit("Input file must have at least 3 columns: unit label, feature label, count, matching the customized column names (case insensitive) --unit_label, --feature_label, and --key/--train_on")

    # To be safe
    model_f = args.output+".model.p"
    if os.path.exists(model_f) and not args.overwrite:
        sys.exit("Model file exists, use --overwrite to allow overwriting")

    # Set up model
    fix_scaling = args.fix_scaling
    factor_header = [str(x) for x in range(K)]
    lda = LDA(n_components=K, learning_method='online', batch_size=b_size, total_samples = args.N, learning_offset = args.tau, learning_decay = args.kappa, doc_topic_prior = args.alpha, n_jobs = args.thread, random_state=seed, verbose = args.verbose)
    lda.shift_log_transform_scale = fix_scaling

    # Genes to use
    with gzip.open(args.feature, 'rt') as rf:
        fheader = rf.readline().strip().split('\t')
    fheader = [x.lower() for x in fheader]
    feature=pd.read_csv(args.feature, sep='\t', skiprows=1, names=fheader, dtype={gene_key:str, key:int})
    feature = feature[feature[key] >= args.min_ct_per_feature]
    feature.sort_values(by=key,ascending=False,inplace=True)
    feature.drop_duplicates(subset=gene_key,keep='first',inplace=True)
    feature_kept = list(feature[gene_key].values)

    # If use prior
    if os.path.isfile(args.prior):
        prior = pd.read_csv(args.prior, sep='\t', header=0)
        if "gene" not in prior.columns:
            sys.exit("ERROR: prior file must have a column named 'gene'")
        if prior.shape[1] != K + 1:
            logging.warn(f"Number of factors in --prior file does not match --nFactor ({K}), will use all factors in the prior file")
        K = prior.shape[1] - 1
        prior = prior[prior.gene.isin(feature_kept)]
        feature_kept = list(prior.gene.values)
        prior.drop(columns=['gene'], inplace=True)
        factor_header = list(prior.columns)
        M = len(feature_kept)
        prior = np.array(prior).T # K x M
        if args.rescale_prior > 0:
            w = prior.sum(axis=1)
            target_w = w * (args.N * args.rescale_prior / w.sum())
            prior = normalize(prior, norm='l1', axis=1) * target_w.reshape((-1, 1))
            print("Scaled prior")
            print(prior.sum(axis = 1).round(2))
        init_latent_vars(lda, n_features = M, gamma = prior)
        mt = prior.sum(axis =1)
        mt = " ".join([f"{x:.2e}" for x in mt])
        logging.info(f"Read prior for global parameters. Prior magnitude: {mt}")

    ft_dict = {x:i for i,x in enumerate( feature_kept ) }
    M = len(ft_dict)
    logging.info(f"Start fitting model ... {M} genes will be used")

    epoch = 0
    n_unit = 0
    chunksize=100000 if args.debug else 2000000
    while epoch < args.epoch:
        reader = pd.read_csv(gzip.open(args.input, 'rt'), \
                sep='\t',chunksize=chunksize, skiprows=1, names=header, \
                usecols=[unit_key,gene_key,train_on], dtype=adt)
        batch_obj =  UnitLoader(reader, ft_dict, train_on, \
            batch_id_prefix=args.epoch_id_length, \
            min_ct_per_unit=args.min_ct_per_unit,
            unit_id=unit_key,unit_attr=[])
        if args.shift_log_transform and n_unit == 0 and fix_scaling < 0:
            N = batch_obj.read_one_epoch();
            if N == 0:
                sys.exit("ERROR: no unit read from input file")
            unit_sum = batch_obj.mtx.sum(axis=1)
            qt_sum = np.quantile(unit_sum, q = .98)
            fix_scaling = np.log(1. / qt_sum + 1)
            logging.info(f"Shift log transform scaling: {fix_scaling:.3e} based on .98 quantile of unit sum {qt_sum}")
            batch_obj.mtx = normalize(batch_obj.mtx, norm='l1', axis=1)
            batch_obj.mtx.data = np.log(batch_obj.mtx.data + 1) / fix_scaling
            _ = lda.partial_fit(batch_obj.mtx)
            n_unit += N
            if args.verbose or args.debug:
                logl = lda.score(batch_obj.mtx) / N
                logging.info(f"Epoch {epoch}, finished {n_unit} units. batch logl: {logl:.4f}")
            if args.epoch == 1:
                break
        while batch_obj.update_batch(b_size):
            N = batch_obj.mtx.shape[0]
            if args.verbose or args.debug:
                x1 = np.median(batch_obj.brc[train_on].values)
                x2 = np.mean(batch_obj.brc[train_on].values)
                logging.info(f"Made DGE {N}, median/mean count: {x1:.1f}/{x2:.1f}")
            if args.shift_log_transform:
                batch_obj.mtx = normalize(batch_obj.mtx, norm='l1', axis=1)
                batch_obj.mtx.data = np.log(batch_obj.mtx.data + 1) / fix_scaling
            _ = lda.partial_fit(batch_obj.mtx)
            n_unit += N
            if args.verbose or args.debug:
                logl = lda.score(batch_obj.mtx) / N
                logging.info(f"Epoch {epoch}, finished {n_unit} units. batch logl: {logl:.4f}")
            if len(batch_obj.batch_id_list) > args.epoch:
                break
        if args.epoch_id_length > 0:
            epoch += len(batch_obj.batch_id_list)
        else:
            epoch += 1

    lda.feature_names_in_ = feature_kept
    # Relabel factors based on (approximate) descending abundance
    weight = lda.components_.sum(axis=1)
    ordered_k = np.argsort(weight)[::-1]
    lda.components_ = lda.components_[ordered_k,:]
    lda.exp_dirichlet_component_ = lda.exp_dirichlet_component_[ordered_k,:]
    # Store model
    out_f = args.output + ".model.p"
    pickle.dump( lda, open( out_f, "wb" ) )
    post_mtx = lda.components_.T
    out_f = args.output + ".model_matrix.tsv.gz"
    pd.concat([pd.DataFrame({gene_key: feature_kept}),\
                pd.DataFrame(post_mtx,\
                columns = [str(k) for k in range(K)], dtype='float64')],\
                axis = 1).to_csv(out_f, sep='\t', index=False, float_format='%.4e', compression={"method":"gzip"})

    ### Rerun all units once and store results
    oheader = ["unit",key,"x","y","topK","topP"]+factor_header
    dtp = {'topK':int,key:int,"unit":str}
    dtp.update({x:float for x in ['topP']+factor_header})
    res_f = args.output+".fit_result.tsv.gz"
    nbatch = 0
    logging.info(f"Result file {res_f}")

    ucol = [unit_key,gene_key,key] + unit_attr
    if key != train_on:
        ucol += [train_on]
    reader = pd.read_csv(gzip.open(args.input, 'rt'), \
            sep='\t',chunksize=chunksize, skiprows=1, names=header, \
            usecols=ucol, dtype=adt)
    batch_obj =  UnitLoader(reader, ft_dict, key, \
        batch_id_prefix=args.epoch_id_length, \
        min_ct_per_unit=args.min_ct_per_unit, \
        unit_id=unit_key, unit_attr=unit_attr, train_key=train_on)
    post_count = np.zeros((K, M))
    while batch_obj.update_batch(b_size):
        N = batch_obj.mtx.shape[0]
        if args.shift_log_transform:
            mtx = normalize(batch_obj.mtx, norm='l1', axis=1)
            mtx.data = np.log(mtx.data + 1) / fix_scaling
        else:
            mtx = batch_obj.mtx
        theta = lda.transform(mtx)
        if key != train_on:
            post_count += np.array(theta.T @ batch_obj.test_mtx)
        else:
            post_count += np.array(theta.T @ batch_obj.mtx)
        brc = pd.concat((batch_obj.brc.reset_index(), pd.DataFrame(theta, columns = factor_header)), axis = 1)
        brc['topK'] = np.argmax(theta, axis = 1).astype(int)
        brc['topP'] = np.max(theta, axis = 1)
        brc = brc.astype(dtp)
        logging.info(f"{nbatch}-th batch with {brc.shape[0]} units")
        mod = 'w' if nbatch == 0 else 'a'
        hdr = True if nbatch == 0 else False
        brc[oheader].to_csv(res_f, sep='\t', mode=mod, float_format="%.4e", index=False, header=hdr, compression={"method":"gzip"})
        nbatch += 1
        if args.epoch_id_length > 0 and len(batch_obj.batch_id_list) > 1:
            break

    logging.info(f"Finished ({nbatch})")

    out_f = args.output+".posterior.count.tsv.gz"
    pd.concat([pd.DataFrame({gene_key: feature_kept}),\
            pd.DataFrame(post_count.T, dtype='float64',\
                            columns = [str(k) for k in range(K)])],\
            axis = 1).to_csv(out_f, sep='\t', index=False, float_format='%.2f', compression={"method":"gzip"})

if __name__ == "__main__":
    lda(sys.argv[1:])



# Visualize pixel level single factor heatmap
# Input file contains only top k factors and probabilities per pixel
# Meant to make use of the indexed input to plot for specified regions quickly
# Would take a huge amount of memory if trying to plot many factors simultaneously in a large region

import sys, os, copy, gc, re, gzip, pickle, argparse, logging, warnings
import numpy as np
import pandas as pd
from scipy.sparse import *
import matplotlib as mpl
import cv2

from ficture.loaders.pixel_factor_loader import BlockIndexedLoader

def plot_pixel_single(_args):

    parser = argparse.ArgumentParser(prog="plot_pixel_single")
    parser.add_argument('--input', type=str, help='')
    parser.add_argument('--output', type=str, help="Output prefix")
    parser.add_argument('--id_list', type=str, nargs="+", help="List of IDs of the factors to plot")
    parser.add_argument('--pcut', type=float, default=1e-2, help="")
    parser.add_argument('--binary_cmap_name', type=str, default="plasma", help="Name of Matplotlib colormap to use for ploting individual factors")

    parser.add_argument('--xmin', type=float, default=-np.inf, help="")
    parser.add_argument('--ymin', type=float, default=-np.inf, help="")
    parser.add_argument('--xmax', type=float, default=np.inf, help="")
    parser.add_argument('--ymax', type=float, default=np.inf, help="")
    parser.add_argument('--org_coord', action='store_true', help="If the input coordinates do not include the offset (if your coordinates are from an existing figure, the offset is already factored in)")
    parser.add_argument('--full', action='store_true', help="Read full input")
    parser.add_argument('--plot_um_per_pixel', type=float, default=1, help="Actual size (um) corresponding to each pixel in the output image")
    parser.add_argument('--all', action="store_true", help="Caution: when set, assume factors are named as 0, 1, ... K-1, where K is defined in the input header. Only use when plotting a small region.")
    parser.add_argument('--debug', action='store_true', help="")

    args = parser.parse_args(_args)
    if len(_args) == 0:
        parser.print_help()
        return

    logging.basicConfig(level= getattr(logging, "INFO", None), format='%(asctime)s %(message)s', datefmt='%I:%M:%S %p')
    if not os.path.exists(os.path.dirname(args.output)):
        os.makedirs(os.path.dirname(args.output))

    loader = BlockIndexedLoader(args.input, args.xmin, args.xmax, args.ymin, args.ymax, args.full, not args.org_coord)
    width = int((loader.xmax - loader.xmin + 1)/args.plot_um_per_pixel)
    height= int((loader.ymax - loader.ymin + 1)/args.plot_um_per_pixel)
    logging.info(f"Image size {height} x {width}")
    if args.debug:
        logging.info(f"{loader.xmin}, {loader.xmax}, {loader.ymin}, {loader.ymax}")

    id_list = args.id_list
    if args.all:
        id_list = [str(k) for k in range(loader.meta['K'])]

    # Read input file
    df = pd.DataFrame()
    for chunk in loader:
        if chunk.shape[0] == 0:
            continue
        chunk['X'] = np.clip(((chunk.X - loader.xmin) / args.plot_um_per_pixel).astype(int),0,width-1)
        chunk['Y'] = np.clip(((chunk.Y - loader.ymin) / args.plot_um_per_pixel).astype(int),0,height-1)
        for k in id_list:
            chunk[k] = np.zeros(chunk.shape[0], dtype=np.float16)
            for i in range(1, loader.meta['TOPK']+1):
                indx = chunk['K'+str(i)].eq(k) & chunk['P'+str(i)].gt(args.pcut)
                chunk.loc[indx, k] = chunk.loc[indx, 'P'+str(i)]
        chunk = chunk.groupby(by=['X','Y']).agg({k:np.mean for k in id_list}).reset_index()
        pmax = chunk.loc[:, id_list].max(axis = 1)
        df = pd.concat([df, chunk.loc[pmax > args.pcut, :]])
        logging.info(f"Reading pixels... {chunk.X.iloc[-1]}, {chunk.Y.iloc[-1]}, {df.shape[0]}")
    df.drop_duplicates(subset=['X','Y'], inplace=True)
    if df.shape[0] == 0:
        sys.exit("ERROR: No pixels found")

    logging.info(f"Read {df.shape[0]} pixels")

    for k in id_list:
        indx = df[k] > args.pcut
        if (df[k] > .5).sum() < 10 and indx.sum() < 100:
            continue
        rgb = np.clip(mpl.colormaps['plasma'](df.loc[indx, k].values)[:,:3]*255,0,255).astype(np.uint8)
        img = np.zeros((height,width,3), dtype=np.uint8)
        for i in range(3):
            img[:, :, i] = coo_array( ( rgb[:, i], (df.loc[indx, 'Y'], df.loc[indx, 'X']) ), shape=(height, width) ).toarray()
        img = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)
        cv2.imwrite(args.output +".F_" +k+".png",img)
        logging.info(f"Made image for {k}")

    logging.info(f"Finished")

if __name__ == "__main__":
    plot_pixel_single(sys.argv[1:])



### Read (randomized) hexagons from file, construct minibatch
import sys, os, gzip, copy, re
import numpy as np
import pandas as pd
from scipy.sparse import coo_array

from ficture.models.lda_minibatch import PairedMinibatch

class UnitLoaderAugmented:

    def __init__(self, reader, ft_dict, key, bkey, batch_id_prefix=0, min_ct_per_unit=1, unit_attr=['x','y']) -> None:
        self.reader = reader
        self.ft_dict = ft_dict
        self.key = key
        self.bkey= bkey
        self.df = pd.DataFrame()
        self.file_is_open = True
        self.batch = None
        self.brc = None
        self.prefix = batch_id_prefix
        self.min_ct_per_unit = min_ct_per_unit
        self.unit_attr = list(unit_attr)
        self.M = max(self.ft_dict.values()) + 1
        self.batch_id_list = set()

    def update_batch(self, bsize):
        n_unit = 0
        if len(self.df) > 0:
            n_unit = len(self.df.unit.unique())
        if not self.file_is_open:
            return 0
        while n_unit <= bsize:
            try:
                chunk = next(self.reader)
            except StopIteration:
                self.file_is_open = False
                break
            self.df = pd.concat([self.df, chunk])
            n_unit = len(self.df.unit.unique())
        left = pd.DataFrame()
        if self.file_is_open:
            last_indx = self.df.unit.iloc[-1]
            left = self.df[self.df.unit == last_indx]
            self.df = self.df[self.df.unit != last_indx]
        self.brc = self.df[['unit']+self.unit_attr].drop_duplicates(subset=['unit'])
        self.brc = self.brc.merge(right = self.df.groupby(by='unit').agg({self.key:"sum", self.bkey:"sum"}).reset_index(), on = 'unit', how = 'inner' )
        self.brc = self.brc[self.brc[self.key] >= self.min_ct_per_unit]
        buffer_weight = self.brc[self.bkey].values / (self.brc[self.bkey].values + self.brc[self.key].values)
        barcode_kept=list(self.brc.unit)
        bt_dict={x:i for i,x in enumerate(barcode_kept)}
        N = len(bt_dict)
        self.df = self.df[self.df.unit.isin(bt_dict) & self.df.gene.isin(self.ft_dict)]
        if self.prefix > 0:
            self.batch_id_list.update(set(self.df.unit.map(lambda x : x[:self.prefix]).values))
        self.batch = PairedMinibatch(\
            mtx_focal = coo_array((self.df[self.key].values, \
                            (self.df.unit.map(bt_dict).values, \
                             self.df.gene.map(self.ft_dict).values) ), \
                            shape=(N, self.M)).tocsr(),
            mtx_buffer = coo_array((self.df[self.bkey].values, \
                            (self.df.unit.map(bt_dict).values, \
                             self.df.gene.map(self.ft_dict).values) ), \
                            shape=(N, self.M)).tocsr(),
            buffer_weight = buffer_weight)
        self.df = copy.copy(left)
        return self.batch.n



class UnitLoader:

    def __init__(self, reader, ft_dict, key, batch_id_prefix=0, min_ct_per_unit=1, unit_id='unit', unit_attr=['x','y'], train_key=None, epoch = 2**15, skip_epoch=[], debug=False) -> None:
        self.reader = reader
        self.ft_dict = ft_dict
        self.key = key
        self.df = pd.DataFrame()
        self.file_is_open = True
        self.mtx = None
        self.brc = None
        self.prefix = batch_id_prefix
        self.epoch = epoch
        self.min_ct_per_unit = min_ct_per_unit
        self.unit_id = unit_id
        self.unit_attr = list(unit_attr)
        self.M = max(self.ft_dict.values()) + 1
        self.debug = debug
        self.batch_id_list = []
        self.skip_epoch = set(skip_epoch)
        self.train_key = key if train_key is None else train_key
        self.test_mtx = None

    def _make_matrix(self):
        self.brc = self.df[['unit']+self.unit_attr].drop_duplicates(subset=['unit'])
        if self.prefix > 0:
            lab = self.brc.unit.str[:self.prefix].unique()
            self.batch_id_list += [x for x in lab if x not in self.batch_id_list]
        self.brc = self.brc.merge(right = self.df.groupby(by='unit').agg({self.train_key:"sum"}).reset_index(), on = 'unit', how = 'inner' )
        if self.key != self.train_key:
            self.brc = self.brc.merge(right = self.df.groupby(by='unit').agg({self.key:"sum"}).reset_index(), on = 'unit', how = 'inner' )
        self.brc = self.brc[self.brc[self.key] >= self.min_ct_per_unit]
        barcode_kept=list(self.brc.unit)
        bt_dict={x:i for i,x in enumerate(barcode_kept)}
        N = len(bt_dict)
        self.df = self.df[self.df.unit.isin(bt_dict)]
        self.mtx = coo_array((self.df[self.train_key].values, \
                            (self.df.unit.map(bt_dict).values, \
                             self.df.gene.map(self.ft_dict).values) ), \
                            shape=(N, self.M)).tocsr()
        if self.key != self.train_key:
            self.test_mtx = coo_array(( \
                (self.df[self.key] - self.df[self.train_key]).values, \
                            (self.df.unit.map(bt_dict).values, \
                             self.df.gene.map(self.ft_dict).values) ), \
                            shape=(N, self.M)).tocsr()
            self.test_mtx.eliminate_zeros()
        return N

    def update_batch(self, bsize):
        n_unit = 0
        if len(self.df) > 0:
            n_unit = len(self.df.unit.unique())
        if len(self.batch_id_list) > self.epoch or not self.file_is_open:
            return 0
        while n_unit <= bsize:
            try:
                chunk = next(self.reader)
            except StopIteration:
                self.file_is_open = False
                break
            chunk.rename(columns={self.unit_id:'unit'}, inplace=True)
            chunk = chunk[chunk.gene.isin(self.ft_dict)]
            if len(self.skip_epoch) > 0:
                lab = chunk.unit.str[:self.prefix]
                chunk = chunk[~lab.isin(self.skip_epoch)]
            if len(chunk) == 0:
                continue
            if self.debug:
                print(f"Read {chunk.shape[0]} lines from file")
            self.df = pd.concat([self.df, chunk])
            n_unit = len(self.df.unit.unique())
        left = pd.DataFrame()
        if self.file_is_open:
            last_indx = self.df.unit.iloc[-1]
            left = self.df[self.df.unit == last_indx]
            self.df = self.df[self.df.unit != last_indx]
        N = self._make_matrix()
        self.df = copy.copy(left)
        return N


    def read_one_epoch(self):
        if not self.file_is_open:
            return 0
        if self.prefix <= 0:
            print(f"UnitLoader::read_one_epoch Will read the whole file")
        left = pd.DataFrame()
        local_epoch_list = []
        if len(self.df > 0) and self.prefix > 0:
            local_epoch_list = list(np.unique([x[:self.prefix] for x in self.df.unit.unique()]) )
        while len(local_epoch_list) < 2:
            try:
                chunk = next(self.reader)
            except StopIteration:
                self.file_is_open = False
                break
            chunk.rename(columns={self.unit_id:'unit'}, inplace=True)
            chunk = chunk[chunk.gene.isin(self.ft_dict)]
            if len(chunk) == 0:
                continue
            if self.prefix > 0:
                lab = np.unique([x[:self.prefix] for x in chunk.unit.unique()])
                local_epoch_list += [x for x in lab if x not in local_epoch_list]
                if len(local_epoch_list) > 1:
                    left = chunk[~chunk.unit.str[:self.prefix].eq(local_epoch_list[0])]
                    chunk = chunk[chunk.unit.str[:self.prefix].eq(local_epoch_list[0])]
            self.df = pd.concat([self.df, chunk])
            if self.debug:
                print(f"Read {chunk.shape[0]} lines from file")
        N = self._make_matrix()
        self.df = copy.copy(left)
        return N



import sys, os, copy, gc, re, gzip, pickle, argparse, logging, warnings
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

import sklearn.neighbors
from scipy.sparse import coo_array
from datetime import datetime
from sklearn.manifold import MDS

from ficture.utils.utilt import plot_colortable
# from ficture.utils.mds_color_circle import assign_color_mds_circle

def assign_color_mds_line(mtx, cmap_name, weight=None, top_color=None, seed=None):
    # mtx is a K by K similarity/proximity matrix
    assert mtx.shape[0] == mtx.shape[1], "mtx must be square"
    K = mtx.shape[0]
    # weight is a K vector of factor abundance
    if weight is None:
        weight = np.ones(K)
    weight /= weight.sum()
    # The color of the top factor (the one with the largest weight)
    if top_color is None:
        top_color = "#fcd217"
    else:
        match = re.search(r'^#(?:[0-9a-fA-F]{3}){1,2}$', top_color)
        if match is None:
            top_color = "#fcd217"
    # Find the offset to map the top factor to the desired color
    cgrid = 200
    cmtx=plt.get_cmap(cmap_name)(np.arange(cgrid)/cgrid)
    h = top_color.lstrip('#')
    top_color_rgb = [int(h[i:i+2], 16)/255 for i in (0, 2, 4)]
    d = np.abs(cmtx[:, :3] - np.array(top_color_rgb).reshape((1, -1)) ).sum(axis = 1)
    anchor_pos = d.argmin() / cgrid
    anchor_angle = anchor_pos * 2 * np.pi

    mds = MDS(n_components=1, dissimilarity="precomputed", random_state=seed)
    mds_coordinates = mds.fit_transform(mtx).flatten()
    c_order = np.argsort(np.argsort(mds_coordinates))
    w_vec = weight[np.argsort(mds_coordinates)]
    w_vec = np.cumsum(w_vec) - w_vec/2
    normalized_coordinates = np.zeros(K)
    normalized_coordinates[c_order] = w_vec
    angle = normalized_coordinates * 2 * np.pi
    anchor_k = np.argmax(weight)
    angle_shift = angle + (anchor_angle - angle[anchor_k])
    if angle_shift.max() > 2*np.pi:
        angle_shift -= np.pi * 2
    angle_shift[angle_shift < 0] = 2 * np.pi + angle_shift[angle_shift < 0]
    c_pos = angle_shift / np.pi / 2
    return c_pos

def choose_color(_args):

    parser = argparse.ArgumentParser()
    parser.add_argument('--input', type=str, help='')
    parser.add_argument('--output', type=str, help='')
    parser.add_argument('--cmap_name', type=str, default="turbo", help="Name of Matplotlib colormap to use (better close to a circular colormap)")
    parser.add_argument('--top_color', type=str, default="#fcd217", help="HEX color code for the top factor")
    parser.add_argument('--even_space', action='store_true', help="Evenly space the factors on the circle")
    parser.add_argument('--annotation', type=str, default = '', help='')
    parser.add_argument('--circle', action='store_true', help="")
    parser.add_argument('--seed', type=int, default=-1, help='')
    args = parser.parse_args(_args)

    if len(_args) == 0:
        parser.print_help()
        return

    ## obtain seed if not provided
    seed = args.seed
    if seed <= 0:
        seed = int(datetime.now().timestamp()) % 2147483648
    np.random.seed(seed)

    factor_name = {}
    if os.path.isfile(args.annotation):
        with open(args.annotation) as f:
            for line in f:
                x = line.strip().split('\t')
                factor_name[x[0]] = x[1]

    cmap_name = args.cmap_name
    if args.cmap_name not in plt.colormaps():
        cmap_name = "turbo"

    df = pd.read_csv(args.input, sep='\t', header=0)
    df.rename(columns = {"X":"x","Y":"y"},inplace=True)
    header = df.columns
    factor_header = []
    for x in header:
        y = re.match(r'^[A-Za-z]*_*(\d+)$', x)
        if y:
            factor_header.append(y.group(0))
    K = len(factor_header)
    N = df.shape[0]

    # Factor abundance (want top factors to have more distinct colors)
    if args.even_space:
        weight=None
    else:
        weight = df.loc[:, factor_header].sum(axis = 0).values
        weight = weight**(1/2)
        weight /= weight.sum()
        weight = np.clip(weight, .2/K, 1)
        weight /= weight.sum()

    # Find neearest neighbors
    bt = sklearn.neighbors.BallTree(df.loc[:, ["x", "y"]])
    dist, indx = bt.query(df.loc[:, ["x", "y"]], k = 7, return_distance=True)
    r_indx = np.array([i for i,v in enumerate(indx) for y in range(len(v))], dtype=int)
    c_indx = indx.reshape(-1)
    dist = dist.reshape(-1)
    nn = dist[dist > 0].min()
    mask = (dist < nn + .5) & (dist > 0)
    r_indx = r_indx[mask]
    c_indx = c_indx[mask]
    # Compute spatial similarity
    Sig = coo_array((np.ones(len(r_indx)), (r_indx, c_indx)), shape=(N, N)).tocsr()
    W = np.array(df.loc[:, factor_header])
    mtx = W.T @ Sig @ W
    # Translate into a symmetric similarity measure
    # Large values in mtx indicate close proximity, to be mapped to distinct colors
    np.fill_diagonal(mtx, 0)
    mtx /= mtx.sum(axis = 1)
    mtx = mtx + mtx.T

    if args.circle:
        # Assign color using MDS with circular constraint
        from ficture.utils.mds_color_circle import assign_color_mds_circle
        c_pos = assign_color_mds_circle(mtx, cmap_name, weight=weight, top_color=args.top_color, seed=seed)
    else:
        c_pos = assign_color_mds_line(mtx, cmap_name, weight=weight, top_color=args.top_color, seed=seed)

    # # Assign color using MDS with circular constraint
    # c_pos = assign_color_mds_circle(mtx, cmap_name, weight=weight, top_color=args.top_color, seed=seed)

    spectral_offset = .05 # avoid extremely dark colors
    c_pos = (c_pos - c_pos.min()) / (c_pos.max() - c_pos.min()) * (1 - spectral_offset) + spectral_offset

    c_rank = np.argsort(np.argsort(c_pos))
    cmtx = plt.get_cmap(cmap_name)(c_pos) # K x 4
    df = pd.DataFrame({"Name":np.arange(K).astype(str), "Color_index":c_rank})
    df = pd.concat([df, pd.DataFrame(cmtx[:, :3], columns=["R", "G", "B"])], axis=1)
    cdict = {k:cmtx[k,:] for k in range(K)}
    if len(factor_name) > 0:
        df["Annotation"] = df.Name.map(factor_name)
        cdict = {factor_name[str(k)]:cmtx[k,:] for k in range(K)}

    # Output RGB table
    f = args.output + ".rgb.tsv"
    df.to_csv(f, sep='\t', index=False)

    # Plot color bar
    fig = plot_colortable(cdict, "Factor label", sort_colors=False, ncols=4)
    f = args.output + ".cbar.png"
    fig.savefig(f, format="png", transparent=True)


if __name__ == "__main__":
    choose_color(sys.argv[1:])




# ### zz - Previous approach

# linear = MDS(n_components=1, dissimilarity="precomputed",normalized_stress='auto').fit_transform(mtx).squeeze()
# c_order = np.argsort(linear)
# c_rank  = np.argsort(c_order)

# # Uniform
# c_pos = np.linspace(.1, .9, K)

# # Weighted
# weight = df.loc[:, factor_header].sum(axis = 0)
# weight = weight**(1/4)
# weight /= weight.sum()
# weight = weight[c_order]
# c_up = np.cumsum(weight)
# c_down =  np.concatenate([[0], np.cumsum(weight[:-1]) ])
# c_pos = (c_up + c_down) / 2
# c_pos = (c_pos - c_pos[0]) / (c_pos[-1] - c_pos[0]) * .8 + .1

# cmtx = plt.get_cmap(cmap_name)(c_pos) # K x 4
# cdict = {i:cmtx[x,:] for i,x in enumerate(c_rank)}
# df = pd.DataFrame({"Name":range(K), "Color_index":c_rank})
# df = pd.concat([df, pd.DataFrame(cmtx[c_rank, :3], columns=["R", "G", "B"]) ], axis=1)

# # Allocate colors to major factors first
# weight = df.loc[:, factor_header].sum(axis = 0)
# weight /= weight.sum()
# w_order = np.argsort(weight)[::-1]
# w_cdf = np.cumsum(weight[w_order])
# k = np.arange(K)[w_cdf > .9][0]
# major_k = sorted(w_order[:k+1] )
# minor_k = sorted(w_order[k+1:] )
# c_rank_major = np.argsort( np.argsort(linear[major_k]) )
# c_rank_minor = np.argsort( np.argsort(linear[minor_k]) )
# c_pos = np.linspace(args.major_lower, args.major_upper, len(major_k))
# cmtx_major = plt.get_cmap(cmap_name)(c_pos)[c_rank_major, :]
# cdict = {x:cmtx_major[i,:] for i,x in enumerate(major_k)}
# c_pos = np.linspace(0, 1, len(minor_k))
# cmtx_minor = plt.get_cmap(cmap_name)(c_pos)[c_rank_minor, :]
# cdict.update({x:cmtx_minor[i,:] for i,x in enumerate(minor_k)} )
# cmtx = np.vstack([cmtx_major, cmtx_minor])
# df = pd.DataFrame({"Name":list(major_k)+list(minor_k), "Color_index":c_rank})
# df = pd.concat([df, pd.DataFrame(cmtx[:, :3], columns=["R", "G", "B"]) ], axis=1)
# df.sort_values(by = "Name", inplace=True)

# # This is wrong
# cmap = plt.get_cmap(cmap_name, K)
# cmtx = np.array([cmap(i) for i in range(K)] )
# c_order = np.argsort(linear)
# df = pd.DataFrame({"Name":range(K), "Color_index":c_order,\
#         "R":cmtx[c_order, 0], "G":cmtx[c_order, 1], "B":cmtx[c_order, 2]})
# cdict = {i:cmtx[x,:] for i,x in enumerate(c_order)}



import sys, os, copy, gc, re, gzip, pickle, argparse, logging, warnings
import numpy as np
import pandas as pd
from random import shuffle
from scipy.sparse import *

import matplotlib as mpl
import matplotlib.pyplot as plt
from PIL import Image

from ficture.utils.hexagon_fn import *
from ficture.utils.utilt import plot_colortable

def plot_hexagon(_args):

    parser = argparse.ArgumentParser()
    parser.add_argument('--input', type=str, help='')
    parser.add_argument('--output', type=str, help='Output prefix')
    # parser.add_argument('--fill_range', type=float, default=0, help="um")
    parser.add_argument('--hex_radius', type=float, default=-1, help="um")
    parser.add_argument('--hex_diam', type=float, default=-1, help="um")
    parser.add_argument('--batch_size', type=float, default=500, help="")
    parser.add_argument("--tif", action='store_true', help="Store as 16-bit tif instead of png")
    parser.add_argument('--scale', type=float, default=-1, help="")
    parser.add_argument('--origin', type=int, default=[0,0], help="{0, 1} x {0, 1}, specify how to orient the image w.r.t. the coordinates. (0, 0) means the lower left corner has the minimum x-value and the minimum y-value; (0, 1) means the lower left corner has the minimum x-value and the maximum y-value;")
    parser.add_argument('--category_column', type=str, default='', help='')

    parser.add_argument('--color_table_category_name', type=str, default='Name', help='When --category_column is provided, which column to use as the category name')
    parser.add_argument('--binary_cmap_name', type=str, default="plasma", help="Name of Matplotlib colormap to use for ploting individual factors")
    parser.add_argument('--color_table', type=str, default='', help='Pre-defined color map')
    parser.add_argument('--input_rgb_uint8', action="store_true",help="If input rgb is from 0-255 instead of 0-1")
    parser.add_argument('--cmap_name', type=str, default="turbo", help="Name of Matplotlib colormap, only used when --color_table is not provided")

    parser.add_argument("--plot_fit", action='store_true', help="")
    parser.add_argument('--xmin', type=float, default=-np.inf, help="")
    parser.add_argument('--ymin', type=float, default=-np.inf, help="")
    parser.add_argument('--xmax', type=float, default=np.inf, help="")
    parser.add_argument('--ymax', type=float, default=np.inf, help="")
    parser.add_argument('--plot_um_per_pixel', type=float, default=1, help="Size of the output pixels in um")

    parser.add_argument("--skip_mixture_plot", action='store_true', help="")
    parser.add_argument("--plot_discretized", action='store_true', help="")
    parser.add_argument("--plot_individual_factor", action='store_true', help="")
    parser.add_argument("--debug", action='store_true', help="")

    args = parser.parse_args(_args)
    if len(_args) == 0:
        parser.print_help()
        return

    logging.basicConfig(level= getattr(logging, "INFO", None))
    dt = np.uint16 if args.tif else np.uint8
    kcol = args.category_column
    ccol = args.color_table_category_name
    # radius = args.fill_range/args.plot_um_per_pixel
    radius = args.hex_radius
    if radius <= 0:
        radius = args.hex_diam / np.sqrt(3)
    if radius <= 0:
        sys.exit("ERROR: invalid --hex_radius or --hex_diam")

    # Dangerous way to detect which columns to use as factor loadings
    with gzip.open(args.input, "rt") as rf:
        header = rf.readline().strip().split('\t')
    # Temporary - to be compatible with older input files
    recolumn = {'Hex_center_x':'x', 'Hex_center_y':'y', 'X':'x', 'Y':'y'}
    for i,x in enumerate(header):
        if x in recolumn:
            header[i] = recolumn[x]
    factor_header = []
    categorical = False
    if args.category_column != '':
        if args.category_column not in header:
            sys.exit(f"ERROR: {args.category_column} not found in header")
        categorical = True
        if not os.path.exists(args.color_table):
            sys.exit(f"ERROR: --color_table is required for categorical input")
        color_info = pd.read_csv(args.color_table, sep='\t', header=0, dtype={ccol:str})
        if args.input_rgb_uint8 or color_info[["R","G","B"]].max().max() > 2:
            for c in list("RGB"):
                color_info[c] = color_info[c] / 255
        color_info = color_info[~color_info[ccol].isna()]
        color_idx = {x:i for i,x in enumerate(color_info[ccol].values)}
        cmtx = np.array(color_info.loc[:, ["R","G","B"]])
        K = len(color_idx)
        factor_header = [str(k) for k in range(K)]
        logging.info(f"Use categorical input ({K} categories)")
    else:
        for x in header:
            y = re.match('^[A-Za-z]*_*(\d+)$', x)
            if y:
                factor_header.append([y.group(0), int(y.group(1)) ])
        factor_header.sort(key = lambda x : x[1] )
        factor_header = [x[0] for x in factor_header]
        K = len(factor_header)
        if os.path.exists(args.color_table):
            color_info = pd.read_csv(args.color_table, sep='\t', header=0)
            if args.input_rgb_uint8 or color_info[["R","G","B"]].max().max() > 2:
                for c in list("RGB"):
                    color_info[c] = color_info[c] / 255
            cmtx = np.array(color_info.loc[:, ["R","G","B"]])
        else:
            cmap_name = args.cmap_name
            if args.cmap_name not in plt.colormaps():
                cmap_name = "turbo"
            cmap = plt.get_cmap(cmap_name, K)
            cmtx = np.array([cmap(i) for i in range(K)] )
            indx = np.arange(K)
            shuffle(indx)
            cmtx = cmtx[indx, ]
            cmtx = cmtx[:, :3]
            cdict = {k:cmtx[k,:] for k in range(K)}
            # Plot color bar separately
            fig = plot_colortable(cdict, "Factor label", sort_colors=False, ncols=4)
            f = args.output + ".cbar"
            fig.savefig(f, format="png")
            logging.info(f"Set up color map for {K} factors")

    # Read data
    adt={x:float for x in ["x", "y"]+factor_header}
    adt[kcol] = str
    df = pd.DataFrame()
    for chunk in pd.read_csv(gzip.open(args.input, 'rt'), sep='\t', \
        chunksize=1000000, skiprows=1, names=header, dtype=adt):
        if args.scale > 0:
            chunk.x = chunk.x / args.scale
            chunk.y = chunk.y / args.scale
        chunk = chunk[(chunk.y > args.ymin) & (chunk.y < args.ymax)]
        chunk = chunk[(chunk.x > args.xmin) & (chunk.x < args.xmax)]
        # Assume input are hexagons, recover hexagon coordinates
        chunk['x_hex'], chunk['y_hex'] = pixel_to_hex(chunk[['x', 'y']].values, radius, 0, 0)
        if categorical: # Add dummy columns
            chunk[kcol] = chunk[kcol].map(color_idx).astype(int)
            for k in range(K):
                chunk[str(k)] = chunk[kcol].eq(k).astype(int)
        df = pd.concat([df, chunk])

    df.drop_duplicates(inplace=True,subset=['x_hex','y_hex'])
    df.index = list(zip(df.x_hex, df.y_hex))

    x_min = args.xmin
    y_min = args.ymin
    x_max = args.xmax
    y_max = args.ymax
    if args.plot_fit or np.isinf(args.xmin):
        x_min = np.max([x_min, df.x.min()-radius])
    if args.plot_fit or np.isinf(args.ymin):
        y_min = np.max([y_min, df.y.min()-radius])
    if args.plot_fit or np.isinf(args.xmax):
        x_max = np.min([x_max, df.x.max()+radius])
    if args.plot_fit or np.isinf(args.ymax):
        y_max = np.min([y_max, df.y.max()+radius])

    x_indx_min = int(x_min / args.plot_um_per_pixel)
    y_indx_min = int(y_min / args.plot_um_per_pixel)
    x_indx_max = int(x_max / args.plot_um_per_pixel) + 1
    y_indx_max = int(y_max / args.plot_um_per_pixel) + 1

    hsize = x_indx_max - x_indx_min
    wsize = y_indx_max - y_indx_min

    N0 = df.shape[0]
    hsize_um = hsize * args.plot_um_per_pixel
    wsize_um = wsize * args.plot_um_per_pixel
    logging.info(f"Read {N0} pixels in region {hsize_um} x {wsize_um}")


    # Make images
    wst = y_indx_min
    wed = y_indx_max
    wstep = np.max([10, int(args.batch_size)])
    print(wsize, wstep)

    pts = np.zeros((0, 2), dtype=int)
    pts_indx = []
    st = wst
    while st < wed:
        ed = min([st + wstep, wed])
        logging.info(f"Filling pixels {st} - {ed} / {wed}")

        st_um = st / args.plot_um_per_pixel
        ed_um = ed / args.plot_um_per_pixel
        block = df.index[(df.y > st_um - radius) & (df.y < ed_um + radius)]

        mesh = np.meshgrid(np.arange(hsize), np.arange(st, ed))
        nodes = np.array(list(zip(*(dim.flat for dim in mesh))), dtype=float)
        nodes[:, 0] += x_indx_min
        nodes *= args.plot_um_per_pixel # original scale
        x, y = pixel_to_hex(nodes, radius, 0, 0)
        hex_xy = list(zip(x, y))
        indx = [i for i,x in enumerate(hex_xy) if x in block]
        nodes = nodes[indx, :]
        pts_indx += [hex_xy[i] for i in indx]
        nodes /= args.plot_um_per_pixel
        nodes[:, 0] -= x_indx_min
        nodes[:, 1] -= y_indx_min
        pts = np.vstack((pts, nodes.astype(int) ) )
        st = ed
        if args.debug:
            break

    # Note: PIL default origin is upper-left
    pts[:,0] = np.clip(hsize - pts[:, 0], 0, hsize-1)
    pts[:,1] = np.clip(pts[:, 1], 0, wsize-1)
    if args.origin[0] > 0:
        pts[:,0] = hsize - 1 - pts[:, 0]
    if args.origin[1] > 0:
        pts[:,1] = wsize - 1 - pts[:, 1]

    logging.info(f"Start constructing RGB image")

    if not args.skip_mixture_plot:
        rgb_mtx = np.clip(np.around(np.array(df.loc[pts_indx,factor_header]) @\
                                    cmtx * 255),0,255).astype(dt)
        img = np.zeros( (hsize, wsize, 3), dtype=dt)
        for r in range(3):
            img[:, :, r] = coo_array((rgb_mtx[:, r], (pts[:,0], pts[:,1])),\
                shape=(hsize, wsize), dtype = dt).toarray()
        if args.tif:
            img = Image.fromarray(img, mode="I;16")
        else:
            img = Image.fromarray(img)

        outf = args.output
        outf += ".tif" if args.tif else ".png"
        img.save(outf)
        logging.info(f"Made fractional image\n{outf}")

    if args.plot_discretized:
        kvec = np.array(df.loc[pts_indx,factor_header]).argmax(axis = 1)
        cmtx = np.clip(np.around(cmtx * 255), 0, 255).astype(dt)
        img = np.zeros( (hsize, wsize, 3), dtype=dt)
        for r in range(3):
            img[:, :, r] = coo_array((cmtx[kvec, r], (pts[:,0], pts[:,1])),\
                shape=(hsize, wsize), dtype = dt).toarray()
        if args.tif:
            img = Image.fromarray(img, mode="I;16")
        else:
            img = Image.fromarray(img)

        outf = args.output + ".top"
        outf += ".tif" if args.tif else ".png"
        img.save(outf)
        logging.info(f"Made hard threshold image\n{outf}")

    if args.plot_individual_factor:
        if args.binary_cmap_name not in plt.colormaps():
            args.binary_cmap_name = "plasma"
        v = np.array(df.loc[pts_indx, factor_header].sum(axis = 0) )
        u = np.argsort(-v)
        for k in u:
            v = np.clip(df.loc[pts_indx, factor_header[k]].values,0,1)
            rgb_mtx = np.clip(mpl.colormaps[args.binary_cmap_name](v)[:,:3]*255,0,255).astype(dt)
            img = np.zeros( (hsize, wsize, 3), dtype=dt)
            for r in range(3):
                img[:, :, r] = coo_array((rgb_mtx[:, r], (pts[:,0], pts[:,1])),\
                    shape=(hsize, wsize), dtype = dt).toarray()
            if args.tif:
                img = Image.fromarray(img, mode="I;16")
            else:
                img = Image.fromarray(img)
            outf = args.output + ".F_"+str(k)
            outf += ".tif" if args.tif else ".png"
            img.save(outf)
            logging.info(f"Made factor specific image - {k}\n{outf}")

if __name__ == "__main__":
    plot_hexagon(sys.argv[1:])



import sys, os, gzip, copy, re, logging, warnings
import numpy as np
import pandas as pd
import subprocess as sp

class BlockIndexedLoader:

    def __init__(self, input, xmin = -np.inf, xmax = np.inf, ymin = -np.inf, ymax = np.inf, full = False, offseted = True, filter_cmd = "", idtype={}, chunksize=1000000) -> None:
        self.meta = {}
        self.header = []
        nheader = 0
        with gzip.open(input, 'rt') as rf:
            for line in rf:
                if line[0] != "#":
                    break
                nheader += 1
                if line[:2] == "##":
                    wd = line[(line.rfind("#")+1):].strip().split(';')
                    wd = [[y.strip() for y in x.strip().split("=")] for x in wd]
                    for v in wd:
                        if v[1].lstrip('-+').isdigit():
                            self.meta[v[0]] = int(v[1])
                        elif v[1].replace('.','',1).lstrip('-+').isdigit():
                            self.meta[v[0]] = float(v[1])
                        else:
                            self.meta[v[0]] = v[1]
                else:
                    self.header = line[(line.rfind("#")+1):].strip().split('\t')
        logging.basicConfig(level= getattr(logging, "INFO", None), format='%(asctime)s %(message)s', datefmt='%I:%M:%S %p')
        logging.info("Read header %s", self.meta)

        if np.isinf(xmin) and np.isinf(xmax) and np.isinf(ymin) and np.isinf(ymax):
            full = True
        self.xmin = xmin if offseted else xmin - self.meta['OFFSET_X']
        self.xmax = xmax if offseted else xmax - self.meta['OFFSET_X']
        self.ymin = ymin if offseted else ymin - self.meta['OFFSET_Y']
        self.ymax = ymax if offseted else ymax - self.meta['OFFSET_Y']
        # Input reader
        dty={'BLOCK':str, 'X':int, 'Y':int}
        if 'TOPK' in self.meta:
            dty.update({f"K{k+1}" : str for k in range(self.meta['TOPK']) })
            dty.update({f"P{k+1}" : float for k in range(self.meta['TOPK']) })
        if len(idtype) > 0:
            dty.update(idtype)
        self.file_is_open = True
        self.xmin = max(xmin, 0)
        self.xmax = min(xmax, self.meta["SIZE_X"])
        self.ymin = max(ymin, 0)
        self.ymax = min(ymax, self.meta["SIZE_Y"])
        if full:
            self.reader = pd.read_csv(input,sep='\t',skiprows=nheader,chunksize=chunksize,names=self.header, dtype=dty)
        else:
            # Translate target region to index
            if self.meta['BLOCK_AXIS'] == "Y":
                block = [int(x / self.meta['BLOCK_SIZE']) for x in [self.ymin, self.ymax - 1] ]
                pos_range = [int(x*self.meta['SCALE']) for x in [self.xmin, self.xmax]]
            else:
                block = [int(x / self.meta['BLOCK_SIZE']) for x in [self.xmin, self.xmax - 1] ]
                pos_range = [int(x*self.meta['SCALE']) for x in [self.ymin, self.ymax]]
            block = np.arange(block[0], block[1]+1) * self.meta['BLOCK_SIZE']
            query = []
            pos_range = '-'.join([str(x) for x in pos_range])
            for i,b in enumerate(block):
                query.append( str(b)+':'+pos_range )

            cmd = " ".join( ["tabix", input] + query )
            if filter_cmd != "":
                cmd = cmd + " | " + filter_cmd
            logging.info(cmd)
            process = sp.Popen(cmd, stdout=sp.PIPE, stderr=sp.STDOUT, shell=True)
            self.reader = pd.read_csv(process.stdout,sep='\t',chunksize=chunksize,names=self.header, dtype=dty)

    def __iter__(self):
        return self

    def __next__(self):
        if not self.file_is_open:
            raise StopIteration
        try:
            chunk = next(self.reader)
        except StopIteration:
            self.file_is_open = False
            raise StopIteration
        chunk['X']=chunk.X/self.meta['SCALE']
        chunk['Y']=chunk.Y/self.meta['SCALE']
        drop_index = chunk.index[(chunk.X<self.xmin)|(chunk.X>self.xmax)|\
                                 (chunk.Y<self.ymin)|(chunk.Y>self.ymax)]
        return chunk.drop(index=drop_index)



import copy
import numpy as np
import pandas as pd
import sklearn.neighbors
from scipy.sparse import *
import matplotlib as mpl

#########################################################
############# Whole data in memory, write image by row
#########################################################
class ImgRowIterator:
    def __init__(self, pts, mtx, radius, verbose=500, chunksize=500):
        self.pts = pts
        self.N = self.pts.shape[0]
        self.width, self.height = self.pts.max(axis = 0) + 1
        self.current = -1
        self.mtx = mtx
        self.dt = mtx.dtype
        self.verbose = verbose
        self.radius = radius
        self.chunksize = chunksize
        print(f"Image size (w x h): {self.width} x {self.height}")
        self.buffer_lower = self.pts[:, 1].min()
        self.buffer_upper = self.buffer_lower + self.chunksize
        indx = (self.pts[:, 1] >= self.buffer_lower) & (self.pts[:, 1] <= self.buffer_upper)
        self.buffer_index = np.arange(self.N)[indx]
        while len(self.buffer_index) < 2 and self.buffer_upper < self.height:
            self.buffer_upper += self.chunksize
            indx = (self.pts[:, 1] >= self.buffer_lower) & (self.pts[:, 1] <= self.buffer_upper)
            self.buffer_index = np.arange(self.N)[indx]
        assert len(self.buffer_index) > 1, "Input coordinates are ouside input range"
        if self.radius >= 1:
            self.ref = sklearn.neighbors.BallTree(self.pts[self.buffer_index, :])
        print(f"Initialized buffer for block {self.buffer_lower} - {self.buffer_upper}")
        return

    def __iter__(self):
        return self

    def __next__(self):
        self.current += 1
        if self.current % self.verbose == 0:
            print(f"{self.current}/{self.height}")
        if self.current >= self.height:
            raise StopIteration
        self.update_buffer()
        nodes = np.array([[i, self.current] for i in range(self.width)])
        out = np.zeros(self.width * 3, dtype = self.dt)
        if self.radius < 1:
            iv = np.arange(self.N)[self.pts[:, 1] == self.current]
            iu = self.pts[iv, 0]
        else:
            dv, iv = self.ref.query(nodes, k = 1)
            indx = dv[:, 0] < self.radius
            iv = self.buffer_index[iv[indx, 0]]
            iu = np.arange(self.width)[indx]
        if len(iv) == 0:
            return out
        for c in range(3):
            out[iu*3+c] = self.mtx[iv, c]
        return out

    def update_buffer(self):
        if self.radius < 1:
            return
        if self.current < self.buffer_upper - self.radius - 1:
            return
        if self.buffer_upper >= self.height:
            return
        self.buffer_lower = self.current - self.radius - 1
        self.buffer_upper = self.buffer_lower + self.chunksize
        indx = (self.pts[:, 1] >= self.buffer_lower) & (self.pts[:, 1] <= self.buffer_upper)
        self.buffer_index = np.arange(self.N)[indx]
        while len(self.buffer_index) < 2 and self.buffer_upper < self.height:
            self.buffer_upper += self.chunksize
            indx = (self.pts[:, 1] >= self.buffer_lower) & (self.pts[:, 1] <= self.buffer_upper)
            self.buffer_index = np.arange(self.N)[indx]
        if len(self.buffer_index) > 0:
            self.ref = sklearn.neighbors.BallTree(self.pts[self.buffer_index, :])
            return





#########################################################
############# Stream in data, write image by row
#########################################################
class ImgRowIterator_stream:
    def __init__(self, reader, w, h, cmtx,\
                 xmin = 0, ymin = 0, pixel_size = 1, \
                 verbose=500, dtype=np.uint8, plot_top=0):
        self.reader = reader
        self.cmtx = cmtx
        self.xmin = xmin
        self.ymin = ymin
        self.xmax = xmin + w
        self.ymax = ymin + h
        self.width  = int((self.xmax - self.xmin) / pixel_size) + 1
        self.height = int((self.ymax - self.ymin) / pixel_size) + 1
        self.pixel_size = pixel_size
        self.K = cmtx.shape[0]
        self.buffer_y = -1
        self.current = -1
        self.dtype = dtype
        self.verbose = verbose
        self.plot_top = plot_top
        self.file_is_open = True
        self.feature_header = [str(k) for k in range(self.K)]
        self.data_header = ["x", "y"] + self.feature_header
        self.pts = pd.DataFrame([], columns = ["x", "y"])
        self.mtx = np.zeros((0, 3))
        self.leftover = pd.DataFrame([], columns = self.data_header)
        while self.buffer_y < 0 and self.file_is_open:
            self.file_is_open = self.read_chunk()
        if self.buffer_y < 0:
            print("Input does not contain pixels in range")
            return
        print(f"Image size (w x h): {self.width} x {self.height}")
        y0, y1 = self.pts.y.min(), self.pts.y.max()
        N = self.pts.shape[0]
        print(f"Current buffer: ({y0}, {y1}), {N}")
        return

    def __iter__(self):
        return self

    def __next__(self):
        self.current += 1
        if self.current % self.verbose == 0:
            print(f"{self.current}/{self.height}")
        if self.current >= self.height:
            raise StopIteration
        while self.buffer_y < self.current and self.file_is_open:
            # Read more data
            self.file_is_open = self.read_chunk()
        iv = np.arange(self.pts.shape[0])[self.pts.y.eq(self.current)]
        out = np.zeros(self.width * 3, dtype = self.dtype)
        if len(iv) == 0:
            return out
        for c in range(3):
            for i in iv:
                out[self.pts.x.iloc[i]*3+c] = self.mtx[i, c]
        return out

    def read_chunk(self):
        try:
            chunk = next(self.reader)
        except StopIteration:
            print(f"Reach the end of file")
            return 0
        if chunk.y.min() > self.ymax:
            print(f"Read all pixels in range")
            return 0
        chunk = chunk[(chunk.x > self.xmin) & (chunk.x < self.xmax) &\
                      (chunk.y > self.ymin) & (chunk.y < self.ymax)]
        if chunk.shape[0] == 0:
            return 1
        chunk.x -= self.xmin
        chunk.y -= self.ymin
        # Translate into image pixel coordinates
        chunk['x'] = np.round(chunk.x.values / self.pixel_size, 0).astype(int)
        chunk['y'] = np.round(chunk.y.values / self.pixel_size, 0).astype(int)
        # Concatenate with leftover
        chunk = pd.concat([self.leftover.loc[:, self.data_header],\
                           chunk.loc[:, self.data_header]])
        # Save the last row (incomplete) for later
        indx = chunk.y.eq(chunk.y.max())
        self.leftover = copy.copy(chunk.loc[indx, self.data_header])
        if np.sum(indx) == chunk.shape[0]: # Need to read more
            return 1
        chunk = chunk.loc[~indx, :]
        # Collapse
        chunk = chunk.groupby(by = ["x", "y"]).agg({\
                      x:np.mean for x in self.feature_header }).reset_index()
        self.pts = chunk.loc[:, ["x", "y"]]
        self.pts.x = np.clip(self.pts.x, 0, self.width-1)
        self.pts.y = np.clip(self.pts.y, 0, self.height-1)
        self.buffer_y = self.pts.y.max()
        if self.plot_top:
            N = self.pts.shape[0]
            self.mtx = coo_array((np.ones(N,dtype=self.dtype), (range(N),\
                np.array(chunk.loc[:, self.feature_header]).argmax(axis = 1))),\
                shape=(N, self.K)).toarray()
            self.mtx = np.clip(np.around(self.mtx @ self.cmtx * 255),0,255).astype(self.dtype)
        else:
            self.mtx = np.clip(np.around(np.array(\
                    chunk.loc[:, self.feature_header]) @ self.cmtx * 255),\
                    0, 255).astype(self.dtype)
        return 1








#########################################################
############# Stream in data, write image by row
#########################################################
class ImgRowIterator_stream_singlechannel:
    def __init__(self, reader, w, h, key, \
                cmap='plasma', xmin = 0, ymin = 0, cutoff = .05,\
                pixel_size = 1, verbose = 500, dtype = np.uint8):
        self.reader = reader
        self.cmap = cmap
        self.key = key
        self.xmin = xmin
        self.ymin = ymin
        self.xmax = xmin + w
        self.ymax = ymin + h
        self.width  = int((self.xmax - self.xmin) / pixel_size) + 1
        self.height = int((self.ymax - self.ymin) / pixel_size) + 1
        self.pixel_size = pixel_size
        self.buffer_y = -1
        self.current = -1
        self.dtype = dtype
        self.cutoff = cutoff
        self.verbose = verbose
        self.file_is_open = True
        self.data_header = ["x", "y", self.key]
        self.pts = pd.DataFrame([], columns = ["x", "y"])
        self.mtx = np.zeros((0, 3))
        self.leftover = pd.DataFrame([], columns = self.data_header)
        while self.buffer_y < 0 and self.file_is_open:
            self.file_is_open = self.read_chunk()
        if self.buffer_y < 0:
            print("Input does not contain pixels in range")
            return
        print(f"Image size (w x h): {self.width} x {self.height}")
        if self.cmap not in mpl.colormaps():
            self.cmap = "plasma"
        y0, y1 = self.pts.y.min(), self.pts.y.max()
        N = self.pts.shape[0]
        print(f"Current buffer: ({y0}, {y1}), {N}")
        return

    def __iter__(self):
        return self

    def __next__(self):
        self.current += 1
        if self.current % self.verbose == 0:
            print(f"{self.current}/{self.height}")
        if self.current >= self.height:
            raise StopIteration
        while self.buffer_y < self.current and self.file_is_open:
            # Read more data
            self.file_is_open = self.read_chunk()
        iv = np.arange(self.pts.shape[0])[self.pts.y.eq(self.current)]
        out = np.zeros(self.width * 3, dtype = self.dtype)
        if len(iv) == 0:
            return out
        for c in range(3):
            for i in iv:
                out[self.pts.x.iloc[i]*3+c] = self.mtx[i, c]
        return out

    def read_chunk(self):
        try:
            chunk = next(self.reader)
        except StopIteration:
            print(f"Reach the end of file")
            return 0
        if chunk.y.min() > self.ymax:
            print(f"Read all pixels in range")
            return 0
        # Crop
        chunk = chunk[(chunk.x > self.xmin) & (chunk.x < self.xmax) &\
                      (chunk.y > self.ymin) & (chunk.y < self.ymax) &\
                      (chunk[self.key] > self.cutoff)]
        if chunk.shape[0] == 0:
            return 1
        chunk.x -= self.xmin
        chunk.y -= self.ymin
        # Translate into image pixel coordinates
        chunk['x'] = np.round(chunk.x.values / self.pixel_size, 0).astype(int)
        chunk['y'] = np.round(chunk.y.values / self.pixel_size, 0).astype(int)
        # Concatenate with leftover
        chunk = pd.concat([self.leftover.loc[:, self.data_header],\
                           chunk.loc[:, self.data_header]])
        # Save the last row (incomplete) for later
        indx = chunk.y.eq(chunk.y.max())
        self.leftover = copy.copy(chunk.loc[indx, self.data_header])
        if np.sum(indx) == chunk.shape[0]: # Need to read more
            return 1
        # Collapse
        chunk = chunk.loc[~indx, :]
        chunk = chunk.groupby(by = ["x", "y"]).agg({\
                              self.key : np.mean }).reset_index()
        self.pts = chunk.loc[:, ["x", "y"]]
        self.pts.x = np.clip(self.pts.x, 0, self.width-1)
        self.pts.y = np.clip(self.pts.y, 0, self.height-1)
        self.buffer_y = self.pts.y.max()
        v = np.clip(chunk[self.key].values,0,1)
        self.mtx = np.clip(mpl.colormaps[self.cmap](v)[:,:3] * 255,\
                           0, 255).astype(self.dtype)
        return 1



from scipy import sparse
import sklearn.preprocessing
import numpy as np
import copy

class minibatch:

    def __init__(self):
        self.doc_pts = 0  # Positions of anchor points
        self.n = 0        # Number of anchor points
        self.N = 0        # Number of pixels
        self.M = 0        # Number of features (vocabulary)
        self.psi = None   # N x n: P(j|i)
        self.phi = None   # N x K: P(j|i)
        self.gamma = None # n x K: P(k|n)
        self.alpha = None # Prior for gamma
        self.ElogO = None # d_psi E_q[log P(Cij)]
        self.ll = 0       # log likelihood
        self.anchor_adj = None # Adjacency matrix of anchor points

    def init_from_matrix(self, mtx, doc_pts, w, psi = None, phi = None, m_gamma = None, anchor_adj = None, features = None, barcodes = None):
        assert sparse.issparse(mtx), "Invalid matrix - must be sparse"
        assert sparse.issparse(w), "Invalid w - must be sparse"

        self.mtx = mtx.tocsr()
        self.N, self.M = mtx.shape
        self.doc_pts = doc_pts
        self.n = self.doc_pts.shape[0]

        assert np.min(w.data) >= 0, "Invalid w - value must be nonnegative"
        assert w.shape == (self.N, self.n), "Inconsistent dimensions of w"
        self.ElogO = copy.copy(w)
        self.ElogO.eliminate_zeros()
        if psi is not None:
            assert sparse.issparse(psi),  "Invalid psi: must be sparse matrix"
            assert psi.shape  == (self.N, self.n), "Inconsistent dimensions of psi"
            self.psi = psi
            self.psi = (self.ElogO != 0).multiply(self.psi)
        else:
            self.psi = copy.copy(self.ElogO)
        self.ElogO.data = np.log(self.ElogO.data)

        if phi is not None:
            assert phi.shape[0] == self.N, "Inconsistent dimensions of phi"
            self.phi = sklearn.preprocessing.normalize(phi, norm='l1', axis=1)

        if m_gamma is not None:
            assert m_gamma.shape[0] == self.n, "Inconsistent dimensions of gamma"
            self.alpha = m_gamma
            self.gamma = m_gamma

        self.anchor_adj = anchor_adj
        if features is None:
            self.feature = list(range(self.M))
        else:
            self.feature = features
        if barcodes is None:
            self.barcode = list(range(self.N))
        else:
            self.barcode = barcodes



import sys, os, argparse, logging, gzip, copy, re, time, warnings, pickle
import numpy as np
import pandas as pd
from sklearn.preprocessing import normalize
from datetime import datetime

from ficture.models.online_slda import OnlineLDA
from ficture.loaders.pixel_loader import PixelMinibatch

def slda_decode(_args):

    parser = argparse.ArgumentParser(prog="slda_decode")
    # Innput and output info
    parser.add_argument('--input', type=str, help='')
    parser.add_argument('--model', type=str, help='')
    parser.add_argument('--output', type=str, help='')
    parser.add_argument('--anchor', type=str, help='')
    parser.add_argument('--anchor_in_um', action='store_true')
    parser.add_argument('--feature', type=str, default='', help='')

    # Data realted parameters
    parser.add_argument('--mu_scale', type=float, default=26.67, help='Coordinate to um translate')
    parser.add_argument('--key', type=str, default = 'gn', help='gt: genetotal, gn: gene, spl: velo-spliced, unspl: velo-unspliced')
    parser.add_argument('--batch_id', type=str, default = 'random_index', help='Input has to have a column with this name indicating the minibatch id')
    parser.add_argument('--precision', type=float, default=.1, help='If positive, collapse pixels within X um.')

    # Learning related parameters
    parser.add_argument('--thread', type=int, default=1, help='')
    parser.add_argument('--neighbor_radius', type=float, default=25, help='The radius (um) of each anchor point\'s territory')
    parser.add_argument('--halflife', type=float, default=0.7, help='Control the decay of distance-based weight')
    parser.add_argument('--theta_init_bound_multiplier', type=float, default=.2, help='')
    parser.add_argument('--inner_max_iter', type=int, default=30, help='')
    parser.add_argument('--model_scale', type=float, default=-1, help='')
    parser.add_argument('--seed', type=int, default=-1, help='')

    # Other
    parser.add_argument('--lite_topk_output_pixel', type=int, default=-1)
    parser.add_argument('--lite_topk_output_anchor', type=int, default=-1)
    parser.add_argument('--log', type=str, default = '', help='files to write log to')
    parser.add_argument('--debug', action='store_true')

    args = parser.parse_args(_args)
    if len(_args) == 0:
        parser.print_help()
        return

    if args.log != '':
        try:
            logging.basicConfig(filename=args.log, filemode='a', encoding='utf-8', level=logging.INFO)
        except:
            logging.basicConfig(level= getattr(logging, "INFO", None))
    else:
        logging.basicConfig(level= getattr(logging, "INFO", None))

    if not os.path.exists(args.model):
        sys.exit("ERROR: cannot find model file")
    if not os.path.exists(args.input):
        sys.exit("ERROR: cannot find input file")
    if not os.path.exists(args.anchor):
        sys.exit("ERROR: cannot find anchor file")

    ## obtain seed if not provided
    seed = int(args.seed)
    if seed <= 0:
        seed = int(datetime.now().timestamp()) % 2147483648

    ### Basic parameterse
    mu_scale = 1./args.mu_scale
    radius = args.neighbor_radius
    precision = args.precision
    key = args.key.lower()
    batch_id = args.batch_id.lower()
    chunk_size = 500000

    ### Load model
    factor_names = []
    if args.model.endswith(".tsv.gz") or args.model.endswith(".tsv"):
        # If input is a gzip tsv file
        model = pd.read_csv(args.model, sep='\t')
        factor_names = list(model.columns)[1:]
        feature_kept = model.gene.values
        model = np.array(model.iloc[:,1:]).T
        model = np.clip(model, 0.5, None)
    else:
        # If input is a pickled model object
        try:
            model = pickle.load(open( args.model, "rb" ))
            feature_kept = np.array(model.feature_names_in_)
            model = model.components_
        except:
            sys.exit("ERROR: --model input should be either a tsv file containing gene names and factor profiles, or a pickled model object with at least two attributes, feature_names_in_ and components_, like those defined in scikit-learn LDA model")
    if os.path.isfile(args.feature):
        feature = pd.read_csv(args.feature, sep='\t')
        kept_idx = np.where(np.in1d(feature_kept, feature.gene.values))[0]
        feature_kept = np.array(feature_kept)[kept_idx]
        model = model[:, kept_idx]

    K, M = model.shape
    ft_dict = {x:i for i,x in enumerate( feature_kept ) }
    if args.model_scale > 0:
        model = normalize(model, norm='l1', axis=1) * args.model_scale
    logging.info(f"{M} genes and {K} factors are read from input model")

    slda = OnlineLDA(vocab=feature_kept, K=K, N=1e6, iter_inner=args.inner_max_iter, verbose = 1, seed = seed)
    slda.init_global_parameter(model)
    init_bound = 1./K * args.theta_init_bound_multiplier

    ### Input pixel info (input has to contain certain columns with correct header)
    with gzip.open(args.input, 'rt') as rf:
        oheader = rf.readline().strip().split('\t')
    oheader = [x.lower() if len(x) > 1 else x.upper() for x in oheader]
    input_header = [batch_id,"X","Y","gene",key]
    dty = {x:float for x in ['X','Y']}
    dty[key] = int
    dty.update({x:str for x in [batch_id, 'gene']})
    mheader = [x for x in input_header if x not in oheader]
    if len(mheader) > 0:
        mheader = ", ".join(mheader)
        sys.exit(f"Input misses the following column: {mheader}.")

    pixel_reader = pd.read_csv(args.input, sep='\t', chunksize=chunk_size, \
                skiprows=1, names=oheader, usecols=input_header, dtype=dty)

    pixel_obj = PixelMinibatch(pixel_reader, ft_dict, \
                            batch_id, key, mu_scale, \
                            radius=radius, halflife=args.halflife,\
                            precision=args.precision, thread=args.thread)
    ### anchor info
    pixel_obj.load_anchor(args.anchor, args.anchor_in_um)
    logging.info(f"Read {pixel_obj.grid_info.shape[0]} grid points")
    factor_header = pixel_obj.factor_header

    post_count = np.zeros((K, M))
    n_batch = 0
    while True:
        read_n_batch = pixel_obj.read_chunk(args.thread)
        logging.info(f"Read {read_n_batch} batches ({pixel_obj.dge_mtx.shape})")
        pcount, pixel, anchor  = pixel_obj.run_chunk(slda, init_bound)
        pixel.X = pixel.X.map('{:.2f}'.format)
        pixel.Y = pixel.Y.map('{:.2f}'.format)
        if args.lite_topk_output_pixel > 0 and args.lite_topk_output_pixel <= K:
            X = pixel[factor_header].values
            partial_indices = np.argpartition(X, -args.lite_topk_output_pixel, axis=1)[:, -args.lite_topk_output_pixel:]
            sorted_top_indices = np.argsort(X[np.arange(X.shape[0])[:, None], partial_indices], axis=1)[:, ::-1]
            top_indices = partial_indices[np.arange(partial_indices.shape[0])[:, None], sorted_top_indices]
            top_values = X[np.arange(X.shape[0])[:, None], top_indices]
            for k in range(args.lite_topk_output_pixel):
                pixel[f"K{k+1}"] = top_indices[:, k]
            for k in range(args.lite_topk_output_pixel):
                pixel[f"P{k+1}"] = np.clip(top_values[:, k], 0, 1)
            pixel.drop(columns = factor_header, inplace=True)
        write_mode = 'w' if n_batch == 0 else 'a'
        header_include = True if n_batch == 0 else False
        pixel.to_csv(args.output+".pixel.tsv.gz", sep='\t', index=False, header=header_include, mode=write_mode, float_format="%.2e", compression={"method":"gzip"})
        logging.info(f"Output {pixel.shape[0]} pixels and {anchor.shape[0]} anchors")
        anchor.X = anchor.X.map('{:.2f}'.format)
        anchor.Y = anchor.Y.map('{:.2f}'.format)
        if args.lite_topk_output_anchor > 0 and args.lite_topk_output_anchor <= K:
            X = anchor[factor_header].values
            partial_indices = np.argpartition(X, -args.lite_topk_output_anchor, axis=1)[:, -args.lite_topk_output_anchor:]
            sorted_top_indices = np.argsort(X[np.arange(X.shape[0])[:, None], partial_indices], axis=1)[:, ::-1]
            top_indices = partial_indices[np.arange(partial_indices.shape[0])[:, None], sorted_top_indices]
            top_values = X[np.arange(X.shape[0])[:, None], top_indices]
            for k in range(args.lite_topk_output_anchor):
                anchor[f"K{k+1}"] = top_indices[:, k]
            for k in range(args.lite_topk_output_anchor):
                anchor[f"P{k+1}"] = np.clip(top_values[:, k], 0, 1)
            anchor.drop(columns = factor_header, inplace=True)
        anchor.to_csv(args.output+".anchor.tsv.gz", sep='\t', index=False, header=header_include, mode=write_mode, float_format="%.2e", compression={"method":"gzip"})
        n_batch += read_n_batch
        post_count += pcount
        if not pixel_obj.file_is_open:
            break

    ### Output posterior summaries
    if len(factor_names) == K:
        factor_header = factor_names
    out_f = args.output + ".posterior.count.tsv.gz"
    pd.concat([pd.DataFrame({'gene': feature_kept}),\
            pd.DataFrame(post_count.T, dtype='float64',\
                            columns = factor_header)],\
            axis = 1).to_csv(out_f, sep='\t', index=False, float_format='%.2f', compression={"method":"gzip"})

if __name__ == "__main__":
    slda_decode(sys.argv[1:])



"""

=============================================================
Online Latent Dirichlet Allocation with variational inference
=============================================================

This implementation is modified from Matthew D. Hoffman's onlineldavb code
Link: https://github.com/blei-lab/onlineldavb
"""
# sklearn version: 1.4.1
# Author: Chyi-Kwei Yau
# Author: Matthew D. Hoffman (original onlineldavb implementation)
from numbers import Integral, Real
import sys, os, copy
import numpy as np
import scipy.sparse as sp
from joblib import effective_n_jobs
from scipy.special import gammaln, logsumexp
from sklearn.preprocessing import normalize
from sklearn.base import (
    BaseEstimator,
    ClassNamePrefixFeaturesOutMixin,
    TransformerMixin,
    _fit_context,
)
from sklearn.utils import check_random_state, gen_batches, gen_even_slices
from sklearn.utils._param_validation import Interval, StrOptions
from sklearn.utils.parallel import Parallel, delayed
from sklearn.utils.validation import check_is_fitted, check_non_negative
from sklearn.decomposition._online_lda_fast import (
    _dirichlet_expectation_1d as cy_dirichlet_expectation_1d,
)
from sklearn.decomposition._online_lda_fast import (
    _dirichlet_expectation_2d,
)
from sklearn.decomposition._online_lda_fast import (
    mean_change as cy_mean_change,
)

from ficture.utils.utilt import gen_even_slices_from_list

EPS = np.finfo(float).eps

def _update_doc_distribution(
    X,
    exp_elog_beta,
    doc_topic_prior,
    max_doc_update_iter,
    mean_change_tol,
    cal_sstats,
    random_state,
):
    """E-step: update document-topic distribution.

    Parameters
    ----------
    X               : N x M
    exp_elog_beta   : K x M `exp(E[log(beta)])`.
    doc_topic_prior : float. Prior of document topic distribution `theta`.
    max_doc_update_iter : int. Max number of iterations for updating q(theta)
    mean_change_tol : float. Stopping tolerance for updating q(theta)
    cal_sstats      : bool. Indicate whether to calculate sufficient statistics
    random_state    : RandomState/Generator instance or None

    Returns
    -------
    (gamma, suff_stats) :
        `gamma` is unnormalized topic distribution for each document.
        In the literature, this is `gamma`. we can calculate `E[log(theta)]`
        from it.
        `suff_stats` is expected sufficient statistics for the M-step.
            When `cal_sstats == False`, this will be None.

    """
    is_sparse_x = sp.issparse(X)
    n_samples, n_features = X.shape
    n_topics = exp_elog_beta.shape[0]

    if random_state:
        gamma = random_state.gamma(100.0, 0.01, (n_samples, n_topics)).astype(
            X.dtype, copy=False
        )
    else:
        gamma = np.ones((n_samples, n_topics), dtype=X.dtype)

    # In the literature, this is `exp(E[log(theta)])`
    exp_elog_theta = np.exp(_dirichlet_expectation_2d(gamma))

    # diff on `component_` (only calculate it when `cal_diff` is True)
    suff_stats = (
        np.zeros(exp_elog_beta.shape, dtype=X.dtype) if cal_sstats else None
    )

    if is_sparse_x:
        X_data = X.data
        X_indices = X.indices
        X_indptr = X.indptr

    # These cython functions are called in a nested loop on usually
    # very small arrays (length=n_topics).
    # In that case, finding the appropriate signature of the
    # fused-typed function can be more costly than its execution,
    # hence the dispatch is done outside of the loop.
    ctype = "float" if X.dtype == np.float32 else "double"
    mean_change = cy_mean_change[ctype]
    dirichlet_expectation_1d = cy_dirichlet_expectation_1d[ctype]
    eps = np.finfo(X.dtype).eps

    for idx_d in range(n_samples):
        if is_sparse_x:
            ids = X_indices[X_indptr[idx_d] : X_indptr[idx_d + 1]] # col index for row idx_d
            cnts = X_data[X_indptr[idx_d] : X_indptr[idx_d + 1]]
        else:
            ids = np.nonzero(X[idx_d, :])[0]
            cnts = X[idx_d, ids]

        gamma_d = gamma[idx_d, :]
        # The next one is a copy, since the inner loop overwrites it.
        exp_elog_theta_d = exp_elog_theta[idx_d, :].copy()
        exp_elog_beta_d = exp_elog_beta[:, ids]

        # Iterate between `gamma_d` and `norm_phi` until convergence
        for _ in range(0, max_doc_update_iter):
            last_d = gamma_d

            # The optimal phi_{dwk} is proportional to
            # exp(E[log(theta_{dk})]) * exp(E[log(beta_{dw})]).
            norm_phi = np.dot(exp_elog_theta_d, exp_elog_beta_d) + eps

            gamma_d = exp_elog_theta_d * np.dot(cnts / norm_phi, exp_elog_beta_d.T)
            # Note: adds doc_topic_prior to gamma_d, in-place.
            dirichlet_expectation_1d(gamma_d, doc_topic_prior, exp_elog_theta_d)

            if mean_change(last_d, gamma_d) < mean_change_tol:
                break
        gamma[idx_d, :] = gamma_d

        # Contribution of document d to the expected sufficient
        # statistics for the M step.
        if cal_sstats:
            norm_phi = np.dot(exp_elog_theta_d, exp_elog_beta_d) + eps
            suff_stats[:, ids] += np.outer(exp_elog_theta_d, cnts / norm_phi)

    return (gamma, suff_stats)


class LDA(
    ClassNamePrefixFeaturesOutMixin, TransformerMixin, BaseEstimator
):
    """
    sklearn.discriminant_analysis.LinearDiscriminantAnalysis:
        A classifier with a linear decision boundary, generated by fitting
        class conditional densities to the data and using Bayes' rule.
    """
    _parameter_constraints: dict = {
        "n_components": [Interval(Integral, 0, None, closed="neither")],
        "doc_topic_prior": [None, Interval(Real, 0, 1, closed="both")],
        "topic_word_prior": [None, Interval(Real, 0, 1, closed="both")],
        "learning_method": [StrOptions({"batch", "online"})],
        "learning_decay": [Interval(Real, 0, 1, closed="both")],
        "learning_offset": [Interval(Real, 1.0, None, closed="left")],
        "max_iter": [Interval(Integral, 0, None, closed="left")],
        "batch_size": [Interval(Integral, 0, None, closed="neither")],
        "evaluate_every": [Interval(Integral, None, None, closed="neither")],
        "total_samples": [Interval(Real, 0, None, closed="neither")],
        "perp_tol": [Interval(Real, 0, None, closed="left")],
        "mean_change_tol": [Interval(Real, 0, None, closed="left")],
        "max_doc_update_iter": [Interval(Integral, 0, None, closed="left")],
        "n_jobs": [None, Integral],
        "verbose": ["verbose"],
        "random_state": ["random_state"],
        "intercept": [None, "array-like"],
        "penalty": [None, Interval(Real, 0, 1, closed="left")],
        "debug" : [None, Integral],
    }

    def __init__(
        self,
        n_components=10,
        *,
        doc_topic_prior=None,
        topic_word_prior=None,
        learning_method="batch",
        learning_decay=0.7,
        learning_offset=10.0,
        max_iter=10,
        batch_size=128,
        evaluate_every=-1,
        total_samples=1e6,
        perp_tol=1e-1,
        mean_change_tol=1e-3,
        max_doc_update_iter=100,
        n_jobs=None,
        verbose=0,
        random_state=None,
        intercept=None,
        penalty=0,
        penalize_ambian_only = False,
        debug = 0,
    ):
        self.n_components = n_components
        self.doc_topic_prior = doc_topic_prior
        self.topic_word_prior = topic_word_prior
        self.learning_method = learning_method
        self.learning_decay = learning_decay
        self.learning_offset = learning_offset
        self.max_iter = max_iter
        self.batch_size = batch_size
        self.evaluate_every = evaluate_every
        self.total_samples = total_samples
        self.perp_tol = perp_tol
        self.mean_change_tol = mean_change_tol
        self.max_doc_update_iter = max_doc_update_iter
        self.n_jobs = n_jobs
        self.verbose = verbose
        self.random_state = random_state
        self.intercept = intercept
        self.penalty = penalty
        self.penalize_ambian_only = penalize_ambian_only
        self.debug = debug

    def _init_latent_vars(self, n_features, dtype=np.float64, lambda_=None):
        """Initialize latent variables."""

        self.random_state_ = check_random_state(self.random_state)
        self.n_batch_iter_ = 1
        self.n_iter_ = 0
        self.n_features = n_features

        if self.intercept is not None:
            self.intercept_ = np.asarray(self.intercept).reshape(-1)
            assert len(self.intercept_) == self.n_features
            self.intercept_ /= self.intercept_.sum()
            self.intercept = True
        else:
            self.intercept = False

        if self.doc_topic_prior is None:
            self.doc_topic_prior_ = 1.0 / self.n_components
        else:
            self.doc_topic_prior_ = self.doc_topic_prior

        if self.topic_word_prior is None:
            self.topic_word_prior_ = 1.0 / self.n_components
        else:
            self.topic_word_prior_ = self.topic_word_prior

        init_gamma = 100.0
        init_var = 1.0 / init_gamma
        # In the literature, this is called `lambda`
        if lambda_ is None:
            self.components_ = self.random_state_.gamma(
                init_gamma, init_var, (self.n_components, n_features)
            ).astype(dtype, copy=False)
        else:
            assert lambda_.shape == (self.n_components, n_features)
            self.components_ = lambda_.astype(dtype)

        # In the literature, this is `exp(E[log(beta)])`
        self.exp_elog_beta = np.exp(
            _dirichlet_expectation_2d(self.components_)
        )
        if self.intercept:
            self.exp_elog_beta = np.vstack(
                (self.exp_elog_beta, self.intercept_)
            )

    def _e_step(self, X, cal_sstats, random_init, parallel=None):
        """E-step in EM update.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            Document word matrix.

        cal_sstats : bool
            Parameter that indicate whether to calculate sufficient statistics
            or not. Set ``cal_sstats`` to True when we need to run M-step.

        random_init : bool
            Parameter that indicate whether to initialize document topic
            distribution randomly in the E-step. Set it to True in training
            steps.

        parallel : joblib.Parallel, default=None
            Pre-initialized instance of joblib.Parallel.

        Returns
        -------
        (gamma, suff_stats) :
            `gamma` is unnormalized topic distribution for each
            document. In the literature, this is called `gamma`.
            `suff_stats` is expected sufficient statistics for the M-step.
            When `cal_sstats == False`, it will be None.

        """

        # Run e-step in parallel
        random_state = self.random_state_ if random_init else None

        n_jobs = effective_n_jobs(self.n_jobs)
        if parallel is None:
            parallel = Parallel(n_jobs=n_jobs, verbose=max(0, self.verbose - 1))
        results = parallel(
            delayed(_update_doc_distribution)(
                X[idx_slice, :],
                self.exp_elog_beta,
                self.doc_topic_prior_,
                self.max_doc_update_iter,
                self.mean_change_tol,
                cal_sstats,
                random_state,
            )
            for idx_slice in gen_even_slices(X.shape[0], n_jobs)
        )

        # merge result
        doc_topics, sstats_list = zip(*results)
        gamma = np.vstack(doc_topics)

        if cal_sstats:
            # This step finishes computing the sufficient statistics for the
            # M-step.
            suff_stats = np.zeros(self.exp_elog_beta.shape, dtype=self.exp_elog_beta.dtype)
            for sstats in sstats_list:
                suff_stats += sstats
            suff_stats *= self.exp_elog_beta
        else:
            suff_stats = None

        return (gamma, suff_stats)

    def _em_step(self, X, total_samples, batch_update, parallel=None):
        """EM update for 1 iteration.

        update `_component` by batch VB or online VB.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            Document word matrix.

        total_samples : int
            Total number of documents. It is only used when
            batch_update is `False`.

        batch_update : bool
            Parameter that controls updating method.
            `True` for batch learning, `False` for online learning.

        parallel : joblib.Parallel, default=None
            Pre-initialized instance of joblib.Parallel

        Returns
        -------
        gamma : ndarray of shape (n_samples, n_components)
            Unnormalized document topic distribution.
        """

        # E-step
        _, suff_stats = self._e_step(
            X, cal_sstats=True, random_init=True, parallel=parallel
        )
        if self.intercept:
            suff_stats = suff_stats[:-1, :]

        # M-step
        if batch_update:
            self.components_ = self.topic_word_prior_ + suff_stats
        else:
            # online update
            # In the literature, the weight is `rho`
            weight = np.power(
                self.learning_offset + self.n_batch_iter_, -self.learning_decay
            )
            doc_ratio = float(total_samples) / X.shape[0]
            if self.penalty > 0 and self.n_batch_iter_ > 3:
                korder = np.arange(self.n_components)
                self.random_state_.shuffle(korder)
                w = suff_stats.sum(axis = 1) * self.penalty
                if self.intercept and self.penalize_ambian_only:
                    for k in korder:
                        suff_stats[k, :] = np.clip((suff_stats[k, :] - self.intercept_ * w[k]) / (1-self.penalty), EPS, None)
                else:
                    w0 = self.components_.sum(axis = 1)
                    w0 /= w0.sum()
                    beta0 = normalize(self.components_, axis = 1, norm = 'l1')
                    if self.intercept:
                        w0 = np.concatenate((w0, [1]))
                        w0 /= w0.sum()
                        beta0 = np.vstack((beta0, self.intercept_))
                    for k in korder:
                        lam0 = copy.copy(suff_stats[k, :])
                        for l in range(beta0.shape[0]):
                            if l == k:
                                continue
                            suff_stats[k, :] -= beta0[l, :] * w[k] * w0[l]
                        suff_stats[k, :] = np.clip(suff_stats[k, :] / (1-self.penalty), EPS, None)
                        delta = np.abs(suff_stats[k, :] - lam0).sum()
                        if self.debug:
                            print(f"{k}, {delta * self.penalty / w[k]:.3f}")
            self.components_ *= 1 - weight
            self.components_ += weight * (
                self.topic_word_prior_ + doc_ratio * suff_stats
            )

        # update `component_` related variables
        self.exp_elog_beta = np.exp(
            _dirichlet_expectation_2d(self.components_)
        )
        if self.intercept:
            self.exp_elog_beta = np.vstack(\
                (self.exp_elog_beta, self.intercept_))
        self.n_batch_iter_ += 1
        return

    def _more_tags(self):
        return {
            "preserves_dtype": [np.float64, np.float32],
            "requires_positive_X": True,
        }

    def _check_non_neg_array(self, X, reset_n_features, whom):
        """check X format

        check X format and make sure no negative value in X.

        Parameters
        ----------
        X :  array-like or sparse matrix

        """
        dtype = [np.float64, np.float32] if reset_n_features else self.components_.dtype

        X = self._validate_data( # from sklearn.base.BaseEstimator
            X,
            reset=reset_n_features,
            accept_sparse="csr",
            dtype=dtype,
        )
        check_non_negative(X, whom)

        return X

    @_fit_context(prefer_skip_nested_validation=True)
    def partial_fit(self, X, y=None):
        """Online VB with Mini-Batch update.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            Document word matrix.

        y : Ignored
            Not used, present here for API consistency by convention.

        Returns
        -------
        self
            Partially fitted estimator.
        """
        first_time = not hasattr(self, "components_")

        X = self._check_non_neg_array(
            X, reset_n_features=first_time, whom="LatentDirichletAllocation.partial_fit"
        )
        n_samples, n_features = X.shape
        batch_size = self.batch_size

        # initialize parameters or check
        if first_time:
            self._init_latent_vars(n_features, dtype=X.dtype)

        if n_features != self.components_.shape[1]:
            raise ValueError(
                "The provided data has %d dimensions while the model was trained with feature size %d."
                % (n_features, self.components_.shape[1])
            )

        n_jobs = effective_n_jobs(self.n_jobs)
        idx_randomize = np.arange(n_samples)
        self.random_state_.shuffle(idx_randomize)
        with Parallel(n_jobs=n_jobs, verbose=max(0, self.verbose - 1)) as parallel:
            for idx_slice in gen_even_slices_from_list(idx_randomize, n_samples // batch_size):
                self._em_step(
                    X[idx_slice, :],
                    total_samples=self.total_samples,
                    batch_update=False,
                    parallel=parallel,
                )

        return self

    @_fit_context(prefer_skip_nested_validation=True)
    def fit(self, X, y=None):
        """Learn model for the data X with variational Bayes method.

        When `learning_method` is 'online', use mini-batch update.
        Otherwise, use batch update.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            Document word matrix.

        y : Ignored
            Not used, present here for API consistency by convention.

        Returns
        -------
        self
            Fitted estimator.
        """
        X = self._check_non_neg_array(
            X, reset_n_features=True, whom="LatentDirichletAllocation.fit"
        )
        n_samples, n_features = X.shape
        max_iter = self.max_iter
        evaluate_every = self.evaluate_every
        learning_method = self.learning_method

        batch_size = self.batch_size

        # initialize parameters
        self._init_latent_vars(n_features, dtype=X.dtype)
        # change to perplexity later
        last_bound = None
        n_jobs = effective_n_jobs(self.n_jobs)
        with Parallel(n_jobs=n_jobs, verbose=max(0, self.verbose - 1)) as parallel:
            for i in range(max_iter):
                idx_randomize = np.arange(n_samples)
                self.random_state_.shuffle(idx_randomize)
                if learning_method == "online":
                    n_batch = n_samples // batch_size
                    for idx_slice in gen_even_slices_from_list(idx_randomize, n_batch):
                        self._em_step(
                            X[idx_slice, :],
                            total_samples=n_samples,
                            batch_update=False,
                            parallel=parallel,
                        )
                else:
                    # batch update
                    self._em_step(
                        X, total_samples=n_samples, batch_update=True, parallel=parallel
                    )

                # check perplexity
                if evaluate_every > 0 and (i + 1) % evaluate_every == 0:
                    doc_topics_distr, _ = self._e_step(
                        X, cal_sstats=False, random_init=False, parallel=parallel
                    )
                    bound, ll = self._perplexity_precomp_distr(
                        X, doc_topics_distr, sub_sampling=False
                    )
                    if self.verbose:
                        print(
                            "iteration: %d of max_iter: %d, perplexity: %.4f"
                            % (i + 1, max_iter, bound)
                        )
                    if last_bound and abs(last_bound - bound) < self.perp_tol:
                        break
                    last_bound = bound
                elif self.verbose:
                    print("iteration: %d of max_iter: %d" % (i + 1, max_iter))
                self.n_iter_ += 1

        # calculate final perplexity value on train set
        doc_topics_distr, _ = self._e_step(
            X, cal_sstats=False, random_init=False, parallel=parallel
        )
        self.bound_, self.ll_ = self._perplexity_precomp_distr(
            X, doc_topics_distr, sub_sampling=False
        )

        return self

    def _unnormalized_transform(self, X):
        """Transform data X according to fitted model.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            Document word matrix.

        Returns
        -------
        gamma : ndarray of shape (n_samples, n_components)
            Document topic distribution for X.
        """
        gamma, _ = self._e_step(X, cal_sstats=False, random_init=False)

        return gamma

    def transform(self, X):
        """Transform data X according to the fitted model.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            Document word matrix.

        Returns
        -------
        gamma : ndarray of shape (n_samples, n_components)
            Document topic distribution for X.
        """
        check_is_fitted(self)
        X = self._check_non_neg_array(
            X, reset_n_features=False, whom="LatentDirichletAllocation.transform"
        )
        gamma = self._unnormalized_transform(X)
        gamma /= gamma.sum(axis=1)[:, np.newaxis]
        return gamma

    def _approx_bound(self, X, gamma, sub_sampling):
        """Estimate the variational bound.

        Estimate the variational bound over "all documents" using only the
        documents passed in as X. Since log-likelihood of each word cannot
        be computed directly, we use this bound to estimate it.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            Document word matrix.

        gamma : ndarray of shape (n_samples, n_components)
            Document topic distribution. In the literature, this is called
            gamma.

        sub_sampling : bool, default=False
            Compensate for subsampling of documents.
            It is used in calculate bound in online learning.

        Returns
        -------
        score : float

        """

        def _loglikelihood(prior, distr, dirichlet_distr):
            # calculate log-likelihood
            size = distr.shape[1]
            score = np.sum((prior - distr) * dirichlet_distr)
            score += np.sum(gammaln(distr) - gammaln(prior))
            score += np.sum(gammaln(prior * size) - gammaln(np.sum(distr, 1)))
            return score

        is_sparse_x = sp.issparse(X)
        n_samples, n_components = gamma.shape
        n_features = self.components_.shape[1]
        score = 0

        Elog_theta = _dirichlet_expectation_2d(gamma)
        Elog_beta = _dirichlet_expectation_2d(self.components_)
        if self.intercept:
            Elog_beta = np.vstack((Elog_beta, np.log(self.intercept_)))

        if is_sparse_x:
            X_data = X.data
            X_indices = X.indices
            X_indptr = X.indptr

        # E[log p(docs | theta, beta)]
        for idx_d in range(0, n_samples):
            if is_sparse_x:
                ids = X_indices[X_indptr[idx_d] : X_indptr[idx_d + 1]]
                cnts = X_data[X_indptr[idx_d] : X_indptr[idx_d + 1]]
            else:
                ids = np.nonzero(X[idx_d, :])[0]
                cnts = X[idx_d, ids]
            temp = (
                Elog_theta[idx_d, :, np.newaxis] + Elog_beta[:, ids]
            )
            norm_phi = logsumexp(temp, axis=0)
            score += np.dot(cnts, norm_phi)
        elogl = score

        # compute E[log p(theta | alpha) - log q(theta | gamma)]
        score += _loglikelihood(self.doc_topic_prior_, gamma, Elog_theta)

        # Compensate for the subsampling of the population of documents
        if sub_sampling:
            doc_ratio = float(self.total_samples) / n_samples
            score *= doc_ratio

        # E[log p(beta | eta) - log q (beta | lambda)]
        if self.intercept:
            scale = self.components_.sum()
            lambda_ = np.vstack((self.components_, self.intercept_ * scale))
            score += _loglikelihood(self.topic_word_prior_, lambda_, Elog_beta)
        else:
            score += _loglikelihood(self.topic_word_prior_, self.components_, Elog_beta)

        return score, elogl

    def score(self, X, y=None):
        """Calculate approximate log-likelihood as score.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            Document word matrix.

        y : Ignored
            Not used, present here for API consistency by convention.

        Returns
        -------
        score : float
            Use approximate bound as score.
        """
        check_is_fitted(self)
        X = self._check_non_neg_array(
            X, reset_n_features=False, whom="LatentDirichletAllocation.score"
        )

        gamma = self._unnormalized_transform(X)
        score, ll = self._approx_bound(X, gamma, sub_sampling=False)
        return score, ll

    def _perplexity_precomp_distr(self, X, gamma=None, sub_sampling=False):
        """Calculate approximate perplexity for data X with ability to accept
        precomputed gamma

        Perplexity is defined as exp(-1. * log-likelihood per word)

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            Document word matrix.

        gamma : ndarray of shape (n_samples, n_components), \
                default=None
            Document topic distribution.
            If it is None, it will be generated by applying transform on X.

        Returns
        -------
        score : float
            Perplexity score.
        """
        if gamma is None:
            gamma = self._unnormalized_transform(X)
        else:
            n_samples, n_components = gamma.shape
            if n_samples != X.shape[0]:
                raise ValueError(
                    "Number of samples in X and gamma do not match."
                )

            if n_components != self.n_components:
                raise ValueError("Number of topics does not match.")

        current_samples = X.shape[0]
        bound, ll = self._approx_bound(X, gamma, sub_sampling)

        if sub_sampling:
            word_cnt = X.sum() * (float(self.total_samples) / current_samples)
        else:
            word_cnt = X.sum()
        perword_bound = bound / word_cnt

        return np.exp(-1.0 * perword_bound), ll/word_cnt

    def perplexity(self, X, sub_sampling=False):
        """Calculate approximate perplexity for data X.

        Perplexity is defined as exp(-1. * log-likelihood per word)

        .. versionchanged:: 0.19
           *gamma* argument has been deprecated and is ignored
           because user no longer has access to unnormalized distribution

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            Document word matrix.

        sub_sampling : bool
            Do sub-sampling or not.

        Returns
        -------
        score : float
            Perplexity score.
        """
        check_is_fitted(self)
        X = self._check_non_neg_array(
            X, reset_n_features=True, whom="LatentDirichletAllocation.perplexity"
        )
        return self._perplexity_precomp_distr(X, sub_sampling=sub_sampling)

    @property
    def _n_features_out(self):
        """Number of transformed output features."""
        return self.exp_elog_beta.shape[0]
